#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2012-2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys, os
import re
import pickle
from array import array
from enum import Enum

from .debug import (debug, DEBUG_DRAW, DEBUG_DRAWNORMALS, DEBUG_DRAWAXES,
                           DEBUG_NOPICK, DEBUG_NOBEVEL, DEBUG_NOLABEL, DEBUG_NOBLACK,
                           DEBUG_NOCACHE, DEBUG_MODELFAST, DEBUG_CACHEUPDATE)
from .config import MODELS_DIR
from .utils import filebyteorder, epsilon, get_texcoords_range, Vector

try:
    _
except NameError:
    _ = lambda t: t
    

def get_texcoords(vector, normal, minx, maxx, miny, maxy):
    x, y = vector.rotationyx_normal(normal)
    return (x - minx) / (maxx - minx), (y - miny) / (maxy - miny)
    
        
class EmptyModel:
    type = ''
    symbols = ''
    symbolsI = ''
    faces = ''
    default_rotation = (0., 0.)
    
    size = sizes = ()
    blocks = []
    cell_count = 0
    cells_visible_faces = []
    
    def __init__(self, *unused_args):
        pass
        
    def __str__(self):
        return 'no puzzle'
        
    def gl_vertex_data(self, *unused_args):
        return (0, b'', (0, 0, 0, 0, 0, 0), ([], 0, 0, 0), []), [()]
        
        
class Model:
    bevel_width = 0.01 if DEBUG_NOBEVEL else 0.1
    
    cache_index = {}
    cache_data = {}
    cache_vectors = {} if DEBUG_CACHEUPDATE else []
    
    def __init__(self, mtype, size):
        items = self.cache_index['type'][mtype].items()
        self.type = mtype
        for key, value in items:
            setattr(self, key, value)
        self.mformat = _(self.mformat)
        # size for UI and config, sizes for plugins
        self.size = self.cache_index['normsize'][mtype].get(size, self.defaultsize)
        self.sizes, filename = self.cache_index['sizes'][mtype][self.size]
        
        self.load_data(filename)
        
        self.cell_count = len(self.cell_indices)
        self.rotation_indices = {s:i for i, s in enumerate(self.rotation_symbols)}
        self._inplace_rotations = None
        if self.slicesmode == 'all':
            self.ignored_slice = [None for sz in self.sizes]
        elif self.slicesmode == 'mid':
            self.ignored_slice = [(sz-1)//2 for sz in self.sizes]
        elif self.slicesmode == 'last1':
            self.ignored_slice = [(None if a == 1 else sz-1) for a, sz in enumerate(self.sizes)]
        else:
            assert False, self.slicesmode
        
        #XXX: self.beveled_fvertdata generated here from a nested list, but
        #     the idices come from the modeldata files
        self.beveled_fvertdata = self.gl_bevel_faces()
        
    @property
    def inplace_rotations(self):
        if self._inplace_rotations is None:
            # slow and only used in edit mode
            self._inplace_rotations = []
            for cellidx in range(self.cell_count):
                inplace_rotations = []
                for rotsym, blocknum_blocknum2 in self.rotated_position.items():
                    for blocknum, blocknum2 in enumerate(blocknum_blocknum2):
                        if blocknum == blocknum2 == cellidx and rotsym:
                            inplace_rotations.append(rotsym)
                self._inplace_rotations.append(sorted(inplace_rotations, key=len)[:2])
        return self._inplace_rotations
        
    def is_user_slice(self, iaxis, mslice):
        return self.ignored_slice[iaxis] is None or self.ignored_slice[iaxis] != mslice
        
    def __str__(self):
        return self.mformat.format(*self.size)
        
    @classmethod
    def load_index(cls):
        if DEBUG_NOCACHE:
            try:
                from buildlib import modeldata
            except ImportError:
                pass
            else:
                debug('generating model index')
                cls.cache_index = modeldata.pool_create_indexdata()
                return
        indexfilename = os.path.join(MODELS_DIR, 'f00index' if DEBUG_MODELFAST else 'd00index')
        debug('loading model index:', indexfilename)
        with open(indexfilename, 'rb') as indexfile:
            cls.cache_index = pickle.load(indexfile)
            
    @staticmethod
    def read_vectors(datafile):
        vlen = array('I')
        vlen.fromfile(datafile, 1)
        if sys.byteorder != filebyteorder:
            vlen.byteswap()
        vlen = vlen[0]
        vectors = array('f')
        vectors.fromfile(datafile, vlen)
        if sys.byteorder != filebyteorder:
            vectors.byteswap()
        return vectors
        
    @staticmethod
    def isplit3_vectors(vals):
        it = iter(vals)
        for x in it:
            try:
                yield Vector((x, next(it), next(it)))
            except StopIteration:
                raise ValueError('len must be a multiple of 3')
                
    @classmethod
    def _load_data(self, filename):
        if DEBUG_NOCACHE:
            try:
                from buildlib import modeldata
            except ImportError:
                pass
            else:
                print('generating model data:', filename)
                _data, _vectors, *unused = modeldata.pool_create_modelfiledata(filename)
                _vectors = [Vector(v) for v in _vectors]
                return _data, _vectors
        filename = os.path.join(MODELS_DIR, filename)
        debug('loading model data:', filename)
        with open(filename, 'rb') as datafile:
            data = pickle.load(datafile)
            vectors = list(self.isplit3_vectors(self.read_vectors(datafile)))
        return data, vectors
        
    def load_data(self, filename):
        debug('requesting model data:', filename, self.type, self.sizes)
        try:
            data = (self.cache_data[filename] if DEBUG_CACHEUPDATE else self.cache_data)[self.type][self.sizes]
        except KeyError:
            data, vectors = self._load_data(filename)
            if DEBUG_CACHEUPDATE:
                self.cache_data[filename] = data
                self.cache_vectors[filename] = vectors
            else:
                self.__class__.cache_data = data
                self.__class__.cache_vectors = vectors
            data = data[self.type][self.sizes]
        self.vectors = self.cache_vectors[filename] if DEBUG_CACHEUPDATE else self.cache_vectors
        for key, value in data.items():
            setattr(self, key, value)
            
    @classmethod
    def from_string(cls, modelstr):
        if modelstr == '*':
            return '*', '*', (), None
        re_model = r'''^(\w+)
                        (?: \s+(\w+)
                            (?: \s*(\W)
                                \s*(\w+)
                                (?: \s*\3
                                    \s*(\w+)
                            )?)?
                            (?:\s+with
                                \s+(.+)
                        )?)?\s*$'''
        match = re.match(re_model, modelstr, re.X)
        if match is None:
            raise ValueError('Invalid model: ' + modelstr)
        #TODO: len(sizes) > 3
        mtype, width, height, depth, exp = match.group(1, 2, 4, 5, 6)
        if mtype not in cls.cache_index['types']:
            raise ValueError('Unknown model type %r' % mtype)
        def convert_if_int(value):
            if value is None:
                return value
            try:
                return int(value)
            except ValueError:
                return value
        sizes = tuple(convert_if_int(s) for s in (width, height, depth))
        return modelstr, mtype, sizes, exp
        
    def norm_symbol(self, sym):
        try:
            return self.normal_rotation_symbols[sym]
        except KeyError:
            new_sym = ''
            for c in sym:
                try:
                    new_sym = self.normal_rotation_symbols[new_sym+c]
                except KeyError:
                    raise ValueError('invalid symbol:', sym)
            return new_sym
            
    def blocksym_to_blockpos(self, blocksym):
        for i, faces in enumerate(self.cells_visible_faces):
            #FIXME: different blocks can have same visible faces (eg. edges for 4-Cubes)
            if sorted(faces) == sorted(list(blocksym)):
                return i
        raise ValueError('Invalid block symbol:', blocksym)
        
    def blockpos_to_blocksym(self, blockpos, rotsym):
        #TODO: use cells_visible_faces, see blocksym_to_blockpos
        def axisidx_to_sym(axis, idx):
            if idx <= self.sizes[axis] // 2:
                sym = self.symbols[axis]
            else:
                idx = self.sizes[axis]-1 - idx
                sym = self.symbolsI[axis]
            if idx == 0:
                # skip idx for corners
                return sym, self.face_symbolic_to_face_color(sym, rotsym)
            elif idx == 1 and self.sizes[axis] == 3:
                # for size == 3 there is only one edge
                return '', ''
            else:
                # symbol with index to distinguish edge, but no color because the face is not visible
                return sym + str(idx+1), '?'
        #FIXME: only for models with 3 axes, function only used with debug flags
        x, y, z = self.cell_indices[blockpos]
        symx, colorsymx = axisidx_to_sym(0, x)
        symy, colorsymy = axisidx_to_sym(1, y)
        symz, colorsymz = axisidx_to_sym(2, z)
        return symx + symy + symz, colorsymx + colorsymy + colorsymz
        
    def face_symbolic_to_face_color(self, face, rot):
        for k, v in self.face_permutations[rot].items():
            if v == face:
                return k
        assert False, (face, rot)
        
    def sym_to_move(self, sym, mslice=-1):
        try:
            maxis = self.symbols.index(sym)
            return maxis, mslice, False
        except ValueError:
            maxis = self.symbolsI.index(sym)
            return maxis, mslice, True
        
    def normalize_complete_moves(self, moves):
        syms = ''.join((self.symbols if not mdir else self.symbolsI)[maxis] for maxis, unused_mslice, mdir in moves)
        syms = self.norm_symbol(syms)
        for sym in syms:
            yield self.sym_to_move(sym)
        
    def rotate_symbolic(self, axis, rdir, block, sym):
        rsym = (self.symbols if not rdir else self.symbolsI)[axis]
        block = self.rotated_position[rsym][block]
        sym = self.norm_symbol(sym + rsym)
        return block, sym
        
    def rotate_move(self, complete_move, move):
        caxis, unused_cslice, cdir = complete_move
        maxis, mslice, mdir = move
        caxissym = (self.symbols if cdir else self.symbolsI)[caxis]
        maxissym = (self.symbols if not mdir else self.symbolsI)[maxis]
        raxissym = self.face_permutations[caxissym][maxissym]
        raxis, unused_rslice, rdir = self.sym_to_move(raxissym)
        if mdir != rdir:
            mslice = self.sizes[raxis] - 1 - mslice
        return raxis, mslice, rdir
        
    class MoveParallel (Enum):
        NO, LT, COMPAT, GT = range(4)
        
    def is_parallel(self, iaxis1, slice1, iaxis2, slice2):
        for idx in range(self.cell_count):
            pos, rot = idx, ''
            bslice = self.cell_indices[pos][iaxis1]
            if bslice != slice1:
                continue
            pos, rot = self.rotate_symbolic(iaxis1, False, pos, rot)
            
            bslice = self.cell_indices[pos][iaxis2]
            if bslice != slice2:
                continue
            pos, rot = self.rotate_symbolic(iaxis2, False, pos, rot)
            
            bslice = self.cell_indices[pos][iaxis1]
            if bslice == slice1:
                pos, rot = self.rotate_symbolic(iaxis1, True, pos, rot)
                
            bslice = self.cell_indices[pos][iaxis2]
            if bslice == slice2:
                pos, rot = self.rotate_symbolic(iaxis2, True, pos, rot)
                
            if idx != pos or rot != '':
                return False
        else:
            return True
        assert False
        
    def parallel(self, move1, move2):
        if move1.axis == move2.axis:
            if move1.slice == move2.slice:  return self.MoveParallel.COMPAT
            if move1.slice < move2.slice:   return self.MoveParallel.LT
            if move1.slice > move2.slice:   return self.MoveParallel.GT
        if self.is_parallel(move1.axis, move1.slice, move2.axis, move2.slice):
            if move1.axis < move2.axis: 
                return self.MoveParallel.LT
            else:  # move1.axis > move2.axis
                return self.MoveParallel.GT
        return self.MoveParallel.NO
        
    def _get_selected_moves(self, cellidx, vector):
        for iaxis, axis in enumerate(self.axes):
            vda = vector.dot(axis)
            if -epsilon < vda < epsilon:
                continue
            mslice = self.cell_indices[cellidx][iaxis]
            if mslice is not None:
                if not self.is_user_slice(iaxis, mslice):
                    mslice = -1
                yield iaxis, mslice, vda
            
    def _get_other_moves(self, cellidx, vector):
        '''Moves with axes that are not parallel to vector'''
        for iaxis, axis in enumerate(self.axes):
            vca = vector.cross(axis)
            if all(-epsilon < x < epsilon for x in vca):
                continue
            mslice = self.cell_indices[cellidx][iaxis]
            if mslice is not None:
                if not self.is_user_slice(iaxis, mslice):
                    mslice = -1
                yield iaxis, mslice, vca.normalised()
                
    def get_selected_moves2(self, cellidx, vert1, vert2):
        res = sorted(self._get_selected_moves(cellidx, vert2 - vert1), key=lambda m: -abs(m[2]))
        a1, s1, d1 = res[0]
        if len(res) == 1:
            return True, (a1, s1, d1>0)
        a2, s2, d2 = res[1]
        if s1 >= 0 and s2 < 0:
            return True, (a1, s1, d1>0)
        if s1 < 0 and s2 >= 0:
            return True, (a2, s2, d2>0)
        return False, (a1, s1, d1>0,  a2, s2, d2>0)
        
    def get_selected_move_center(self, cellidx, vector):
        moves = ((ma, ms, md) for ma, ms, md in self._get_selected_moves(cellidx, vector) if ms in [0,self.sizes[ma]-1])
        ma, ms, md = max(moves, key=lambda m: abs(m[2]))
        return ma, ms, md>0
        
    @staticmethod
    def triangulate(indices, reverse=False):
        indices = iter(indices)
        first = next(indices)
        prev = next(indices)
        for index in indices:
            yield first
            if reverse:
                yield index
                yield prev
            else:
                yield prev
                yield index
            prev = index
            
    @staticmethod
    def triangulate_barycentric(count, indices, reverse=False):
        if count == 3:
            index1, index2, index3 = indices
            yield index1, (1., 0., 0.)
            index2 = index2, (0., 1., 0.)
            index3 = index3, (0., 0., 1.)
            if reverse:
                yield index3; yield index2
            else:
                yield index2; yield index3
        elif count == 4:
            # +---+  4---3  +-1-+ 
            # | / |  | / |  3 2 1 
            # +---+  1---2  +-3-+ 
            index1, index2, index3, index4 = indices
            index1 = index1, (1., 1., 0.)
            index2 = index2, (0., 1., 0.)
            index3 = index3, (0., 1., 1.)
            index4 = index4, (0., 1., 0.)
            if reverse:
                yield index1; yield index3; yield index2; yield index1; yield index4; yield index3
            else:
                yield index1; yield index2; yield index3; yield index1; yield index3; yield index4
        elif count == 5:
            #   +---+    4---3    +---+ 
            #  /|  /|   /|  /|   2|  /| 
            # + | / |  5 | / |  + 3 3 2 
            #  \|/  |   \|/  |   1|/  | 
            #   +---+    1---2    +-1-+ 
            index1, index2, index3, index4, index5 = indices
            index1 = index1, (0., 1., 1.)
            index2 = index2, (0., 0., 1.)
            index3 = index3, (1., 0., 1.)
            index4 = index4, (1., 0., 1.)
            index5 = index5, (0., 0., 1.)
            if reverse:
                yield index1; yield index3; yield index2; yield index1; yield index4; yield index3
                yield index1; yield index5; yield index4
            else:
                yield index1; yield index2; yield index3; yield index1; yield index3; yield index4
                yield index1; yield index4; yield index5
        elif count == 6:
            #   +---+      1---6      +---+   
            #  /|\  |\    /|\  |\    1|\  |1  
            # + | \ | +  2 | \ | 5  + 3 \ 3 + 
            #  \|  \|/    \|  \|/    2|  \|2  
            #   +---+      3---4      +---+   
            index1, index2, index3, index4, index5, index6 = indices
            index1a = index1, (0., 1., 1.)
            index1b = index1, (0., 1., 1.)
            index2 = index2, (0., 0., 1.)
            index3 = index3, (1., 0., 1.)
            index4 = index4, (1., 0., 1.)
            index5 = index5, (0., 0., 1.)
            index6 = index6, (0., 1., 1.)
            if reverse:
                yield index1a; yield index3; yield index2; yield index1b; yield index4; yield index3
                yield index1b; yield index6; yield index4; yield index4; yield index6; yield index5
            else:
                yield index1a; yield index2; yield index3; yield index1a; yield index3; yield index4
                yield index1a; yield index4; yield index6; yield index4; yield index5; yield index6
        else:
            assert False, count
            
    def pickarrowdir(self, maxis, mdir, normal):
        if mdir:
            return Vector(self.axes[maxis]).cross(normal).normalised()
        else:
            return normal.cross(Vector(self.axes[maxis])).normalised()
            
    def _gl_pick_triangles_pickdata_helper(self, cellidx, symbol, normal, vc, vert1, vert2, pickdata):
        one, masd = self.get_selected_moves2(cellidx, vert1, vert2)
        if one:
            maxis, mslice, mdir = masd
            arrowdir = self.pickarrowdir(maxis, mdir, normal)
            pickdata.append((maxis, mslice, mdir, cellidx, symbol, arrowdir))
        else:
            maxis1, mslice1, mdir1,  maxis2, mslice2, mdir2 = masd
            vert1_scaled = vc*.1 + vert1*.9 - self.vectors[self.cell_centers[cellidx]]
            #d = sqrt(v1² + v2² + v3² - (v1*a1 + v2*a2 + v3*a3)²)
            sdist11_sdist12 = abs(vert1_scaled.dot(self.axes[maxis1])) - abs(vert1_scaled.dot(self.axes[maxis2]))
            if self.reversepick:
                sdist11_sdist12 = -sdist11_sdist12
            if sdist11_sdist12 < 0:
                arrowdir = self.pickarrowdir(maxis1, mdir1, normal)
                pickdata.append((maxis1, mslice1, mdir1, cellidx, symbol, arrowdir))
                arrowdir = self.pickarrowdir(maxis2, mdir2, normal)
                pickdata.append((maxis2, mslice2, mdir2, cellidx, symbol, arrowdir))
            else:
                arrowdir = self.pickarrowdir(maxis2, mdir2, normal)
                pickdata.append((maxis2, mslice2, mdir2, cellidx, symbol, arrowdir))
                arrowdir = self.pickarrowdir(maxis1, mdir1, normal)
                pickdata.append((maxis1, mslice1, mdir1, cellidx, symbol, arrowdir))
        return one
            
    def _gl_pick_triangulate(self, color1, vertsl, normal, mirror_distance):
        for i, verts in enumerate(vertsl):
            for vert in self.triangulate(verts):
                yield color1 + i, vert
        if mirror_distance is not None:
            mirror_offset = normal * mirror_distance
            vertsl = [[v + mirror_offset for v in verts] for verts in vertsl]
            for i, verts in enumerate(vertsl):
                for vert in self.triangulate(verts, reverse=True):
                    yield color1 + i, vert
                    
    def gl_pick_triangles_center(self, cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata):
        # selection_mode: simple, no adjacent visible faces
        maxis, mslice, mdir = self.get_selected_move_center(cellidx, normal)
        color = len(pickdata)
        pickdata.append((maxis, mslice, mdir, cellidx, symbol, None))
        verts = [self.vectors[i] for i in edges_iverts]
        for vert in self.triangulate(verts):
            yield color, vert
        if mirror_distance is not None:
            mirror_offset = normal * mirror_distance
            color = len(pickdata)
            pickdata.append((maxis, mslice, not mdir, cellidx, symbol, None))
            for vert in self.triangulate(verts, reverse=True):
                yield color, vert + mirror_offset
                
    def gl_pick_triangles_gesture(self, cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata):
        # selection_mode: gesture
        moves = list(self._get_other_moves(cellidx, normal))
        color = len(pickdata)
        pickdata.append((None, moves, False, cellidx, symbol, None))
        verts = [self.vectors[i] for i in edges_iverts]
        for vert in self.triangulate(verts):
            yield color, vert
        if mirror_distance is not None:
            mirror_offset = normal * mirror_distance
            for vert in self.triangulate(verts, reverse=True):
                yield color, vert + mirror_offset
            
    def gl_pick_triangles_edge3(self, cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata):
        # selection_mode: simple, face is on an edge block of a tetrahedron
        # the first and second vertex are on the edge of the tetrahedron
        vert1, vert2, vert3 = (self.vectors[i] for i in edges_iverts)
        vc = (vert1+vert2+vert3) / 3
        color1 = len(pickdata)
        one = self._gl_pick_triangles_pickdata_helper(cellidx, symbol, normal, vc, vert1, vert2, pickdata)
        if one:
            vertsl = [[vert1, vert2, vert3]]
        else:
            v12 = (vert1 + vert2) / 2
            vertsl = [[v12, vert2, vert3], [vert1, v12, vert3]]
        yield from self._gl_pick_triangulate(color1, vertsl, normal, mirror_distance)
            
    def gl_pick_triangles_edge6(self, cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata):
        # selection_mode: simple, face is on an edge block and has 6 vertices, rotates 2 slices
        # the first and second vertex are on the edge
        vert1, vert2, vert3, vert4, vert5, vert6 = (self.vectors[i] for i in edges_iverts)
        vc = (vert1+vert2+vert3+vert4+vert5+vert6) / 6
        v12 = (vert1 + vert2) / 2
        v45 = (vert4 + vert5) / 2
        color = len(pickdata)
        one = self._gl_pick_triangles_pickdata_helper(cellidx, symbol, normal, vc, vert1, vert2, pickdata)
        if one:
            vertsl = [[vert1, vert2, vert3, vert4, vert5, vert6]]
        else:
            vertsl = [[v12, vert2, vert3, vert4, v45], [vert1, v12, v45, vert5, vert6]]
        yield from self._gl_pick_triangulate(color, vertsl, normal, mirror_distance)
                
    def gl_pick_triangles_edge4(self, cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata):
        # selection_mode: simple, face is on an edge block, 4 vertices, rotate 1 or 2 slices,
        # the first and second vertex are on the edge
        vert1, vert2, vert3, vert4 = (self.vectors[i] for i in edges_iverts)
        vc = (vert1+vert2+vert3+vert4) / 4
        color = len(pickdata)
        one = self._gl_pick_triangles_pickdata_helper(cellidx, symbol, normal, vc, vert1, vert2, pickdata)
        if one:
            vertsl = [[vert1, vert2, vert3, vert4]]
        else:
            #assert ncolors == 2
            ec12 = (vert1 + vert2) / 2
            ec34 = (vert3 + vert4) / 2
            vertsl = [[ec12, vert2, vert3, ec34], [vert1, ec12, ec34, vert4]]
        yield from self._gl_pick_triangulate(color, vertsl, normal, mirror_distance)
                
    def gl_pick_triangles_corner34(self, cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata):
        # selection_mode: simple, face is on a corner block, 3 or 4 vertices,
        # the second vert is on the corner, the first and third are on the edges
        if len(edges_iverts) == 3:
            vert1, vert2, vert3 = (self.vectors[iv] for iv in edges_iverts)
            vert4 = (vert3 + vert1) / 2
        else:
            vert1, vert2, vert3, vert4 = (self.vectors[iv] for iv in edges_iverts)
        vc = (vert1+vert2+vert3+vert4) / 4
        # first triangle
        color = len(pickdata)
        one = self._gl_pick_triangles_pickdata_helper(cellidx, symbol, normal, vc, vert1, vert2, pickdata)
        if one:
            vertsl = [[vert1, vert2, vert4]]
        else:
            ec12 = (vert1 + vert2) / 2
            vertsl = [[ec12, vert2, vert4], [vert1, ec12, vert4]]
        # second triangle
        one = self._gl_pick_triangles_pickdata_helper(cellidx, symbol, normal, vc, vert2, vert3, pickdata)
        if one:
            vertsl.append([vert2, vert3, vert4])
        else:
            ec23 = (vert2 + vert3) / 2
            vertsl.extend([[ec23, vert3, vert4], [vert2, ec23, vert4]])
        yield from self._gl_pick_triangulate(color, vertsl, normal, mirror_distance)
                    
    def gl_pick_triangles_corner5(self, cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata):
        # selection_mode: simple, face is on a corner block, 5 vertices,
        # the second vert is on the corner, the first and third are on the edges
        vert1, vert2, vert3, vert4, vert5 = (self.vectors[iv] for iv in edges_iverts)
        vc = (vert1+vert2+vert3+vert4+vert5) / 5
        vc13 = (vert1 + vert3) / 2
        ec45 = (vert4 + vert5) / 2
        color1 = len(pickdata)
        one = self._gl_pick_triangles_pickdata_helper(cellidx, symbol, normal, vc, vert1, vert2, pickdata)
        if one:
            vertsl = [[vert1, vert2, ec45, vert5]]
        else:
            vertsl = [[vert1, vert2, vc13], [vert1, vc13, ec45, vert5]]
        one = self._gl_pick_triangles_pickdata_helper(cellidx, symbol, normal, vc, vert2, vert3, pickdata)
        if one:
            vertsl.append([vert2, vert3, vert4, ec45])
        else:
            vertsl.extend([[vc13, vert3, vert4, ec45], [vert2, vert3, vc13]])
        yield from self._gl_pick_triangulate(color1, vertsl, normal, mirror_distance)
        
    def gl_pick_triangles_any(self, cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata):
        # either selection_mode is quadrant or weird model (size==1 in at least one dimension)
        verts = [self.vectors[i] for i in edges_iverts]
        vc = sum(verts, Vector()) / len(verts)
        vert1 = verts[-1]
        for vert2 in verts:
            color1 = len(pickdata)
            one = self._gl_pick_triangles_pickdata_helper(cellidx, symbol, normal, vc, vert1, vert2, pickdata)
            if one:
                vertsl = [[vc, vert1, vert2]]
            else:
                #assert ncolors == 2
                ec = (vert1 + vert2) / 2
                vertsl = [[vc, vert1, ec], [vc, ec, vert2]]
            yield from self._gl_pick_triangulate(color1, vertsl, normal, mirror_distance)
            vert1 = vert2
                
    def gl_pick_triangles(self, selection_mode, mirror_distance, pickdata):
        if DEBUG_NOPICK:
            return
        for unused_fvertdata, (normal, cellidx), (fid, picktype, edges_iverts) in self.beveled_fvertdata:
            if picktype is None:
                continue
            symbol = self.faces[fid]
            if selection_mode == 0:
                yield from self.gl_pick_triangles_any(cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata)
            elif selection_mode == 1:
                if picktype == -1:
                    yield from self.gl_pick_triangles_any(cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata)
                elif picktype == 0:
                    yield from self.gl_pick_triangles_center(cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata)
                elif picktype == 1:
                    yield from self.gl_pick_triangles_edge3(cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata)
                elif picktype == 2:
                    yield from self.gl_pick_triangles_edge4(cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata)
                elif picktype == 3:
                    yield from self.gl_pick_triangles_corner34(cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata)
                elif picktype == 5:
                    yield from self.gl_pick_triangles_corner5(cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata)
                elif picktype == 4:
                    yield from self.gl_pick_triangles_edge6(cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata)
                else:
                    assert False, (selection_mode, picktype)
            elif selection_mode == 2:
                yield from self.gl_pick_triangles_gesture(cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata)
            else:
                assert False, (selection_mode, picktype)
                    
    def bevel_cell_label(self, cidx, c_faces):
        # label faces that replace the old faces
        for hf_face, hf_face_id, hf_verts in c_faces:
            fvertdata1, (fnormal, fcidx), unused = self.beveled_fvertdata[hf_face]
            if cidx != fcidx:
                fnormal = -fnormal
            yield hf_face_id, fnormal, [fvertdata1[v] for v in hf_verts]
                
    def bevel_cell_black(self, cidx, c_faces, c_edges, c_verts):
        fton = lambda fnormal, fcidx: fnormal if cidx == fcidx else -fnormal
        if not DEBUG_NOBEVEL:
            # faces that replace the old verts
            for v, vertdata in c_verts:
                vertdata = [(self.beveled_fvertdata[f][0][v], fton(*self.beveled_fvertdata[f][1])) for f in vertdata]
                yield vertdata
            # faces that replace the old edges
            for f1, f3, v1, v3 in c_edges:
                fvertdata1, finfo, unused = self.beveled_fvertdata[f1]
                normal1 = fton(*finfo)
                fvertdata3, finfo, unused = self.beveled_fvertdata[f3]
                normal3 = fton(*finfo)
                vertdata = [(fvertdata1[v3], normal1), (fvertdata3[v3], normal3),
                            (fvertdata3[v1], normal3), (fvertdata1[v1], normal1)]
                yield vertdata
        # black faces that replace the old faces
        for hf_face, hf_verts in c_faces:
            fvertdata1, finfo, unused = self.beveled_fvertdata[hf_face]
            normal1 = fton(*finfo)
            yield [(fvertdata1[v], normal1) for v in hf_verts]
            
    def gl_label_vertices(self, facesdata, faces, mirror_distance):
        if DEBUG_NOLABEL:
            return
        for faceno, normal, vertices in facesdata:
            fd = faces[faceno]
            color = fd.color
            if fd.imagemode:
                texrange = self.texranges_mosaic[faceno]
            else:
                texrange = get_texcoords_range(vertices, normal)
            texpos = (get_texcoords(v, normal, *texrange) for v in vertices)
            x1,y1,x2,y2 = fd.texrect
            texpos = [(tpx*(x2-x1) + x1, tpy*(y2-y1) + y1) for tpx, tpy in texpos]
            
            for (vertex, tp), baryc in self.triangulate_barycentric(
                                                    len(vertices),
                                                    zip(vertices, texpos)):
                yield vertex, normal, color, tp, baryc
            if mirror_distance is not None:
                offset = normal * mirror_distance
                mirror_normal = -normal
                for (vertex, tp), baryc in self.triangulate_barycentric(
                                                    len(vertices),
                                                    zip(vertices, texpos),
                                                    reverse=True):
                    yield vertex + offset, mirror_normal, color, tp, baryc
                    
    def gl_black_vertices(self, facesdata):
        if DEBUG_NOBLACK:
            return
        for vertices in facesdata:
            for ivertex, normal in self.triangulate(vertices):
                yield ivertex, normal
                
    def gl_bevel_faces(self):
        width = self.bevel_width
        if width == 0:
            def bevel_points(ps):
                return {p: p for p in ps}
        else:
            def bevel_points(ps):
                def bevel_point(point1, point2, point3):
                    point1 = self.vectors[point1]
                    point2 = self.vectors[point2]
                    point3 = self.vectors[point3]
                    nd1 = (point1 - point2).normalised()
                    nd2 = (point3 - point2).normalised()
                    return point2 + (nd1 + nd2) * width / nd1.cross(nd2).length()
                lps = len(ps)
                return {p: bevel_point(ps[(i-1)%lps], p, ps[(i+1)%lps]) for i, p in enumerate(ps)}
        def face_normal(p1, p2, p3, *unused_args):
            p1 = self.vectors[p1]
            p2 = self.vectors[p2]
            p3 = self.vectors[p3]
            return (p3 - p2).cross(p1 - p2).normalised()
        return [(bevel_points(ps), (face_normal(*ps), cidx), (fid, picktype, ps))
                        for cidx, cfd in enumerate(self.facesdata) for fid, picktype, ps in cfd]
                    
    def gl_debug_vertices(self, faces, mirror_distance):
        # yield vertices, colors
        # normals
        if DEBUG_DRAWNORMALS:
            for cidx, (c_faces_label, c_faces_black, c_edges, c_verts) in enumerate(self.blocksdata):
                for faceno, normal, vertices in self.bevel_cell_label(cidx, c_faces_label):
                    fd = faces[faceno]
                    color = fd.color
                    
                    vertex = sum(vertices, Vector()) / len(vertices)
                    yield vertex, [0,0,0]
                    yield vertex+normal, color
                    if mirror_distance is not None:
                        offset = normal * mirror_distance
                        yield vertex+offset, [0,0,0]
                        yield vertex+offset-normal, color
        # axes
        if DEBUG_DRAWAXES:
            for i, axis in enumerate(self.axes):
                symbol = self.symbols[i]
                faceno = self.faces.index(symbol)
                fd = faces[faceno]
                color = fd.color
                vaxis = Vector(axis)*self.sizes[i]
                yield -vaxis, [0,0,0]
                yield -vaxis+axis, color
                yield vaxis, [0,0,0]
                yield vaxis+axis, color
        # coordinates
        x = float(-self.sizes[0])
        y = float(-self.sizes[1])
        z = float(-self.sizes[2])
        yield [x-1, y-1, z,  -x,  y-1,  z], [255, 0, 0,  255, 0, 0]
        yield [x-1, y-1, z,  x-1, -y,   z], [0, 255, 0,  0, 255, 0]
        yield [x-1, y-1, z,  x-1, y-1, -z], [0, 0, 255,  0, 0, 255]
        
    def gl_debug_vertices_2D(self):
        # selection (modelview)
        yield [0, 0, 0,  1, 1, 1], [255, 255, 255,  255, 255, 255]
        # selection (viewport)
        yield [0, 0, 0,  1, 1, 0], [255, 255, 0,  255, 0, 0]
        
    def gl_vertex_data(self, selection_mode, theme, mirror_distance):
        vertices = []
        normals = []
        colors = []
        texpos = []
        baryc = []
        # list of [count vertices, face index]
        cnts_block = []
        faces = [theme.faces[fk] for fk in self.facekeys]
        
        for cidx, (c_faces_label, c_faces_black, c_edges, c_verts) in enumerate(self.blocksdata):
            # bevel
            beveled_facesdata_label = self.bevel_cell_label(cidx, c_faces_label)
            beveled_facesdata_black = self.bevel_cell_black(cidx, c_faces_black, c_edges, c_verts)
            
            cnt = 0
            # label vertices
            for v, n, c, t, b in self.gl_label_vertices(beveled_facesdata_label, faces, mirror_distance):
                vertices += v
                normals += n
                colors += c
                texpos += t
                baryc += b
                cnt += 1
            # bevel, hidden labels
            for v, n in self.gl_black_vertices(beveled_facesdata_black):
                vertices += v
                normals += n
                #TODO: colors and texpos unused
                colors += (0,0,0)
                texpos += (0,0)
                baryc += (0,0,0)
                cnt += 1
            cnts_block.append(cnt)
        
        idx_debug = len(vertices) // 3
        cnt_debug = 0
        if DEBUG_DRAW:
            assert len(cnts_block) == self.cell_count
            assert sum(cnts_block) == idx_debug
            for v, c in self.gl_debug_vertices(faces, mirror_distance):
                vertices += v
                colors += c
            cnt_debug = len(vertices) // 3 - idx_debug
            for v, c in self.gl_debug_vertices_2D():
                vertices += v
                colors += c
            assert len(vertices) == len(colors)
            
        pickvertices, pickcolors, pickdata = self.gl_pick_data(selection_mode, mirror_distance)
        cnt_pick = len(pickvertices) // 3
        
        vertices = array('f', vertices).tobytes()
        normals = array('f', normals).tobytes()
        colors = array('B', colors).tobytes()
        texpos = array('f', texpos).tobytes()
        barycentric = array('f', baryc).tobytes()
        pickvertices = array('f', pickvertices).tobytes()
        pickcolors = array('B', pickcolors).tobytes()
        
        vertexdata = vertices + normals + colors + texpos + barycentric + pickvertices + pickcolors
        debug('GL data: %d bytes' % len(vertexdata))
        
        ptrnormals = len(vertices)
        ptrcolors = ptrnormals + len(normals)
        ptrtexpos = ptrcolors + len(colors)
        ptrbarycentric = ptrtexpos + len(texpos)
        ptrpickvertices = ptrbarycentric + len(barycentric)
        ptrpickcolors = ptrpickvertices + len(pickvertices)
        
        return (    self.cell_count,
                    vertexdata,
                    (ptrnormals, ptrcolors, ptrtexpos, ptrbarycentric, ptrpickvertices, ptrpickcolors),
                    (cnts_block, idx_debug, cnt_debug, cnt_pick),
                    self.rotation_matrices,
               ), pickdata
                
    def gl_pick_data(self, selection_mode, mirror_distance):
        # Pick TRIANGLES vertices
        vertices = []
        colors = []
        pickdata = [()]  # list of (maxis, mslice, mdir, face, center, cellidx, symbol, arrowdir)
        for col, v in self.gl_pick_triangles(selection_mode, mirror_distance, pickdata):
            vertices += v
            colors += ((col>>4) & 0xf0, (col) & 0xf0, (col<<4) & 0xf0)
        assert len(vertices) == len(colors)
        return vertices, colors, pickdata
        
        
empty_model = EmptyModel()
    
    



#-*- coding:utf-8 -*-

#  Copyright © 2009, 2011-2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.



import re
import html

from . import config
from .settings import settings
from .ext import qt  #FIXME: buildlib.utils imports dialogs and the purepython version of qt with PyQt5
from .theme import Theme
from .model import Model
from .application import Meta  #FIXME: Circular import: works because dialogs is imported only inside functions

try:
    _
except NameError:
    _ = lambda t: t
    
    
def linkedtext_to_html(text):
    html = qt.text_to_html(text)
    html = re.sub(r'&lt;((?:http:|https:|text:).*?)\|&gt;', r'<a href="\1">', html)
    html = re.sub(r'&lt;\|&gt;', r'</a>', html)
    return re.sub(r'&lt;((?:http:|https:).*?)&gt;', r'&lt;<a href="\1">\1</a>&gt;', html)
    
    
def init_preferences(sample_buffers):
    shader_names = {'lighting': _('Lighting'), 'simple': _('Simple'),
                    #TRANSLATORS: labels are the colored stickers on the cube
                    'label': _('Labels')}
    facenames = [(_(name), key) for key, name in Model.cache_index['facenames']]
    qt.preferences_dialog(shader_names, sample_buffers, facenames, Theme.textures.get_icons())
        
def get_langname_from_code_func():
    try:
        import icu
    except ImportError:
        print('PyICU is not installed')
        try:
            from .languages import languages
        except ImportError:
            print('No translations for language names')
            return lambda code: None
        else:
            from locale import getlocale, getdefaultlocale
            l = getlocale()[0]
            if l is None:
                l = getdefaultlocale()[0]
            if l is None:
                return lambda code: None
            try:
                languages = languages[l]
            except KeyError:
                try:
                    languages = languages[l.split('_')[0]]
                except KeyError:
                    return lambda code: None
            return lambda code: languages.get(code)
    else:
        def icu_langname_from_code(code):
            iculocale = icu.Locale.createFromName(code)
            lang = iculocale.getLanguage()
            if icu.Locale.createFromName(lang).getDisplayName() == lang:
                return None
            return str(iculocale.getDisplayName()) or None
        return icu_langname_from_code
        
class ModelSelection (metaclass=Meta):
    def __init__(self):
        self.stack = [self.get_modeldata_main]
        self.pagedata = {}
        
    def get_modeldata_main(self):
        defaultindex = -1
        defaultmtype = settings['game.type']
        index = 0
        modeldata_funcs = [None, self.get_modeldata_type1, self.get_modeldata_type2, self.get_modeldata_type3]
        factory_get_modeldata = lambda mtype, func: func and (lambda: func(mtype))
        for mtype in Model.cache_index['types']:
            if mtype == defaultmtype:
                defaultindex = index
            size = Model.cache_index['type'][mtype]['defaultsize']
            text = _(Model.cache_index['type'][mtype]['name'])
            thumbnail = '{}/thumbnails/{}-{}.png'.format(config.UI_DIR, mtype, 'x'.join(str(i) for i in size))
            yield text, thumbnail, mtype, size, factory_get_modeldata(mtype, modeldata_funcs[len(size)])
            index += 1
        yield defaultindex
        yield _('Select Puzzle')
        
    def get_modeldata_type1(self, mtype):
        defaultindex = -1
        defaultsize = settings['games',mtype,'size']
        try:
            defaultsize = Model.cache_index['normsize'][mtype][defaultsize]
        except KeyError:
            defaultsize = Model.cache_index['type'][mtype]['defaultsize']
        index = 0
        for size in sorted(set(Model.cache_index['normsize'][mtype].values())):
            if size == defaultsize:
                defaultindex = index
            text = _(Model.cache_index['type'][mtype]['mformat']).format(*size)
            thumbnail = '{}/thumbnails/{}-{}.png'.format(config.UI_DIR, mtype, 'x'.join(str(i) for i in size))
            yield text, thumbnail, mtype, size, None
            index += 1
        yield defaultindex
        yield _(Model.cache_index['type'][mtype]['name'])
        
    def get_modeldata_type2(self, mtype):
        defaultindex = -1
        defaultsize = Model.cache_index['type'][mtype]['defaultsize']
        index = 0
        factory_get_modeldata = lambda mtype, height: lambda: self.get_modeldata_typeheight2(mtype, height)
        for size in sorted(set(Model.cache_index['normsize'][mtype].values())):
            if size[0] == defaultsize[0] and size[2:] == defaultsize[2:]:
                if size == defaultsize:
                    defaultindex = index
                #TRANSLATORS: 1 slice, 2 slices, 3 slices, ...
                text = ngettext('{} slice','{} slices',size[1]).format(size[1])
                thumbnail = '{}/thumbnails/{}-{}.png'.format(config.UI_DIR, mtype, 'x'.join(str(i) for i in size))
                yield text, thumbnail, mtype, size, factory_get_modeldata(mtype, size[1])
                index += 1
        yield defaultindex
        yield _(Model.cache_index['type'][mtype]['name'])
        
    def get_modeldata_type3(self, mtype):
        defaultindex = -1
        defaultsize = Model.cache_index['type'][mtype]['defaultsize']
        index = 0
        factory_get_modeldata = lambda mtype, height: lambda: self.get_modeldata_typeheight3(mtype, height)
        for size1 in range(1,11):  #FIXME: hardcoded limit
            try:
                size = Model.cache_index['normsize'][mtype][(size1, defaultsize[0], defaultsize[2])]
            except KeyError as e:
                print('hm', e)
                continue
            if size == defaultsize:
                defaultindex = index
            text = ngettext('{} slice','{} slices',size1).format(size1)
            thumbnail = '{}/thumbnails/{}-{}.png'.format(config.UI_DIR, mtype, 'x'.join(str(i) for i in size))
            yield text, thumbnail, mtype, size, factory_get_modeldata(mtype, size1)
            index += 1
        yield defaultindex
        yield _(Model.cache_index['type'][mtype]['name'])
        
    def get_modeldata_typeheight2(self, mtype, height):
        defaultindex = -1
        defaultsize = settings['games',mtype,'size']
        try:
            defaultsize = Model.cache_index['normsize'][mtype][defaultsize]
        except KeyError:
            defaultsize = Model.cache_index['type'][mtype]['defaultsize']
        if defaultsize[1] != height:
            defaultsize = tuple((height if i==1 else s) for i,s in enumerate(defaultsize))
        index = 0
        for size in sorted(set(Model.cache_index['normsize'][mtype].values())):
            if size[1] == height:
                if size == defaultsize:
                    defaultindex = index
                text = _(Model.cache_index['type'][mtype]['mformat']).format(*size)
                thumbnail = '{}/thumbnails/{}-{}.png'.format(config.UI_DIR, mtype, 'x'.join(str(i) for i in size))
                yield text, thumbnail, mtype, size, None
                index += 1
        yield defaultindex
        yield _(Model.cache_index['type'][mtype]['name'])+' – '+ngettext('{} slice','{} slices',height).format(height)
        
    def get_modeldata_typeheight3(self, mtype, height):
        defaultindex = -1
        defaultsize = Model.cache_index['type'][mtype]['defaultsize']
        if defaultsize[1] != height:
            defaultsize = tuple((height if i==1 else s) for i,s in enumerate(defaultsize))
        index = 0
        factory_get_modeldata = lambda mtype, height, width: lambda: self.get_modeldata_typeheightwidth3(mtype, height, width)
        for width in range(1,11):  #FIXME: hardcoded limit
            try:
                size = Model.cache_index['normsize'][mtype][(width, height, defaultsize[2])]
            except KeyError:
                continue
            if size == defaultsize:
                defaultindex = index
            text = _('{}×{} slices').format(width, height)
            thumbnail = '{}/thumbnails/{}-{}.png'.format(config.UI_DIR, mtype, 'x'.join(str(i) for i in size))
            yield text, thumbnail, mtype, size, factory_get_modeldata(mtype, height, width)
            index += 1
        yield defaultindex
        yield _(Model.cache_index['type'][mtype]['name'])+' – '+ngettext('{} slice','{} slices',height).format(height)
        
    def get_modeldata_typeheightwidth3(self, mtype, height, width):
        defaultindex = -1
        defaultsize = settings['games',mtype,'size']
        try:
            defaultsize = Model.cache_index['normsize'][mtype][defaultsize]
        except KeyError:
            defaultsize = Model.cache_index['type'][mtype]['defaultsize']
        if defaultsize[1] != height or defaultsize[0] != width:
            defaultsize = (width, height, defaultsize[2])
        index = 0
        for depth in range(1,11):  #FIXME: hardcoded limit
            try:
                size = Model.cache_index['normsize'][mtype][(width, height, depth)]
            except KeyError:
                continue
            if size == defaultsize:
                defaultindex = index
            text = _(Model.cache_index['type'][mtype]['mformat']).format(*size)
            thumbnail = '{}/thumbnails/{}-{}.png'.format(config.UI_DIR, mtype, 'x'.join(str(i) for i in size))
            yield text, thumbnail, mtype, size, None
            index += 1
        yield defaultindex
        yield _(Model.cache_index['type'][mtype]['name'])+' – '+_('{}×{} slices').format(width, height)
        
    def on_model_get_modeldata(self, pageindex):
        assert pageindex >= 1, (pageindex, self.stack)
        func = self.stack[pageindex-1]
        try:
            items, defaultindex, title = self.pagedata[func]
        except KeyError:
            *items, defaultindex, title = func()
            self.pagedata[func] = [items, defaultindex, title]
        return items, defaultindex, title
    
    def on_model_back(self):
        if len(self.stack) > 1:
            self.stack.pop()
            return True
        return False
    
    def on_convert_to_mtype_size(self, pageindex, index):
        func = self.stack[pageindex-1]
        self.pagedata[func][1] = index
        unused_text, unused_thumbnail, mtype, size, func = self.pagedata[func][0][index]
        if func is None:
            return mtype, size
        else:
            self.stack[pageindex:] = [func]
            return None
    
    
class ProgressDialog:
    def __init__(self):
        self.canceled_ = False
        self.value_max = 10
        self.value = 0
        
    def on_canceled(self):
        self.canceled_ = True
        #XXX: currently not used, so not translation needed
        print(('Canceling operation, please wait'))
        
    def tick(self, step, message=None, value_max=None):
        if message is not None:
            print(message)
        if value_max is not None:
            self.value_max = value_max
            self.value = 0
        if step < 0 or self.value > self.value_max:
            print('  tick')
        else:
            print(' ', self.value, '/', self.value_max)
            self.value += step
        return self.canceled_
        
    def done(self):
        self.canceled_ = False
        
        
class DialogValues:
    @classmethod
    def get_text(self, string):
        return getattr(self, string)()
        
    helpstrings = [
            #TRANSLATORS: The next strings form the text in the help dialog:
            #             Title of the 1st paragraph
            _('Using the mouse to rotate the cube'),
            #TRANSLATORS: Text of the 1st paragraph, followed by bullet list items
            _('Position the mouse cursor over the puzzle and you will see an arrow '
              'that gives you a hint in which direction the slice under the mouse cursor will be rotated.'),
            _('The left mouse button rotates a single slice of the cube in the direction of the arrow.'),
            _('The right mouse button rotates a single slice of the cube against the direction of the arrow.'),
            _('To rotate the whole cube instead of a single slice press the Ctrl key together with the mouse button.'),
            #TRANSLATORS: Title of the 2nd paragraph
            _('Using the keyboard to rotate the cube'),
            #TRANSLATORS: Text of the 2nd paragraph, followed by bullet list items
            _('Make sure the keyboard focus is on the cube area (e.g. click on the background of the cube). '
              'The keys can be configured in the preferences dialog, the default is:'),
            #TRANSLATORS: before the item text is a list of keys
            _('Moves the left, right, upper, down, front or back slice clockwise.'),
            _('Moves a slice couterclockwise.'),
            _('Moves the whole cube.'),
            #TRANSLATORS: Title of the 3rd paragraph
            _('Other keys and buttons'),
            #TRANSLATORS: bullet list items of the 3rd paragraph
            _('Mouse wheel – Zoom in/out'),
            _('Arrow keys, Left mouse button on the background – Changes the direction of looking at the cube.'),
            _('Moves keyboard focus to the sequence editor above the cube area '
              'where you can edit the move sequence in the notation described below. '
              'Hit enter when done.'),
            #TRANSLATORS: Title of the 4th paragraph
            _('Notation for moves'),
            #TRANSLATORS: Text of the 4th paragraph, followed by bullet list items
            _('All moves, however they are made, are displayed progressively above the cube area:'),
            _('Moves the left, right, upper, down, front or back slice clockwise.'),
            _('Moves a slice couterclockwise.'),
            _('Moves the whole cube.'),
            _('Moves the first, second or third slice from left clockwise. '
              'The allowed numbers are in the range from 1 to the count of parallel slices. '
              '"l1" is always the same as "l" '
              'and for the classic 3×3×3-Cube "l2" is the same as "r2-" and "l3" is the same as "r-".'),
            _('You can use a space to separate groups of moves.'),
        ]
    helpformat = '''
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0//EN" "http://www.w3.org/TR/REC-html40/strict.dtd">
<html>
<head>
<meta name="qrichtext" content="1" />
<style type="text/css">p, li {{ white-space: pre-wrap; }}</style>
</head>

<body>

<h3>{}</h3>
<p>{}</p>
<ul>
<li>{}</li>
<li>{}</li>
<li>{}</li>
</ul>

<h3>{}</h3>
<p>{}</p>
<ul>
<li>4,6,8,2,5,0 – {}</li>
<li>Shift+4, … – {}</li>
<li>Ctrl+4, … – {}</li>
</ul>

<h3>{}</h3>
<ul>
<li>{}</li>
<li>{}</li>
<li>Ctrl+L – {}</li>
</ul>

<h3>{}</h3>
<p>{}</p>
<ul>
<li>l, r, u, d, f, b – {}</li>
<li>l-, r-, u-, d-, f-, b- – {}</li>
<li>L, R, U, D, F, B – {}</li>
<li>l1, l2, l3 – {}</li>
<li>{}</li>
</ul>

</body></html>
    '''
    
    @classmethod
    def help_text(self):
        return self.helpformat.format(*[html.escape(s) for s in self.helpstrings])
        
    @staticmethod
    def about_website():
        return '<a href="{}">{}</a>'.format(config.WEBSITE, _('Pybik project website'))
        
    _about_translators = None
    
    @classmethod
    def about_translators(self):
        if self._about_translators is not None:
            return self._about_translators
        html_template = '''<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0//EN" "http://www.w3.org/TR/REC-html40/strict.dtd">
<html><head><meta name="qrichtext" content="1" /><style type="text/css">
p, li {{ white-space: pre-wrap; }}
</style></head>
<body>{}</body></html>'''
        html_p_template = '<p align="center">{}</p>'
        html_language_template1 = '<span style=" font-weight:600;">{}</span>'
        html_language_template2 = '{}'
        html_person_template = '<a href="{}">{}</a>'
        html_p_items = []
        from .translators import translators
        langname_from_code = get_langname_from_code_func()
        import locale
        sortfunc = lambda v: locale.strxfrm(v[0])
        def gentranslators():
            for language, langcode, complete, persons in translators:
                language = langname_from_code(langcode) or language
                yield language, complete, persons
        for language, complete, persons in sorted(gentranslators(), key=sortfunc):
            language = html.escape(language)
            html_items = [(html_language_template1 if complete else html_language_template2).format(language)]
            for name, link in persons:
                name = html.escape(name)
                html_items.append(html_person_template.format(link, name))
            html_p_items.append(html_p_template.format('<br/>'.join(html_items)))
        self._about_translators = html_template.format(''.join(html_p_items))
        return self._about_translators
        
    @staticmethod
    def about_contribute():
        return linkedtext_to_html(config.get_filebug_text() +'\n\n\n'+ _(config.TRANSLATION_TEXT))
        
    @staticmethod
    def about_license_short():
        text = '\n\n'.join((_(config.LICENSE_INFO), _(config.LICENSE_FURTHER)))
        return linkedtext_to_html(text)
        
    @staticmethod
    def about_license_full():
        try:
            with open(config.LICENSE_FILE, 'rt', encoding='utf-8') as license_file:
                text = license_file.read()
            text = linkedtext_to_html(text)
        except Exception as e:
            print('Unable to find license text:', e)
            text = ''
        return text
        
    @staticmethod
    def about_license_notfound():
        text = _(config.LICENSE_NOT_FOUND)
        return linkedtext_to_html(text)
        




# -*- coding: utf-8 -*-

#  Copyright © 2016-2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re


pyxtypes = {'str': 'const QString &', 'bool': 'cbool'}
cpptypes = {'str': 'const QString &', 'object': 'PyObject *'}
pyxrtypes = {'str': 'QString', 'bool': 'cbool'}
cpprtypes = {'str': 'QString', 'object': 'PyObject *'}

cpptype = lambda t: cpptypes.get(t, t)

def bremove(l, value):
    try:
        l.remove(value)
    except ValueError:
        return False
    else:
        return True
        
def parse_cppclass(line):
    match = re.match(r'^class (\w+)\s*(\((.*?)\))?:$', line, re.ASCII)
    base = match.group(3) if match.group(2) else None
    return match.group(1), base
    
def parse_mods(mods):
    mods = mods.split()
    m_mods = mods.pop(0)
    p_mods = []
    c_mods = []
    for mod in mods:
        if mod in ('with', 'gil', 'nogil'):
            p_mods.append(mod)
        else:
            p_mods.append(mod)
            c_mods.append(mod)
    f_mod, *m_mods = m_mods
    p_mods = ' ' + ' '.join(p_mods) if p_mods else ''
    c_mods = ' ' + ' '.join(c_mods) if c_mods else ''
    return f_mod, m_mods, p_mods, c_mods

    
def parse_pfunc(line):
    import inspect
    def mkpyfunc(line):
        exec(line.rsplit(':', 1)[0]+': pass')
        l = locals()
        del l['line']
        assert len(l) == 1, l
        return list(l.values())[0]
    f = mkpyfunc(line)
    s = inspect.signature(f)
    returntype = 'void' if s.return_annotation == inspect._empty else s.return_annotation
    def parameters(s):
        for i, p in enumerate(s.parameters.values()):
            if i == 0 and p.name in ('self', 'this') and p.annotation == inspect._empty:
                continue
            yield ('' if p.annotation == inspect._empty else p.annotation), p.name
    return returntype, f.__name__, list(parameters(s))
    

class PxMakros:
    def PXDFILE(self, indent, line, lineno, *args):
        if line:
            pxdline = self.pxd_indent + indent + line.rstrip(':') + '    #line {}\n'.format(lineno)
            self.pxdf.write(pxdline)
            if self.makrolineno == 0:
                return True
        
    def HFILE(self, indent, line, lineno, *args):
        if line.strip():
            pxhline = indent + line + '    //line {}\n'.format(lineno)
        else:
            pxhline = '\n'
        self.pxhf.write(pxhline)
        return True
        
    def CFILE(self, indent, line, lineno, *args):
        if line.strip():
            line = indent + line + '    //line {}\n'.format(lineno)
        else:
            line = '\n'
        self.pxcf.write(line)
        return True
        
    def CPPCLASS(self, indent, line, lineno, *args, P=True, D=False, Q=False):
        if not line:
            return
        name, base = parse_cppclass(line)
        if base:
            self.HFILE(indent, 'class {} : public {}'.format(name, base), lineno)
            if P:
                self.PYXFILE('cdef cppclass {} ({}):'.format(name, base))
        else:
            self.HFILE(indent, 'class {}'.format(name), lineno)
            if P:
                self.PYXFILE('cdef cppclass {}:'.format(name))
        if D:
            self.PXDFILE(indent, 'cdef cppclass {}:#'.format(name), lineno)
        self.HFILE(indent, '{', lineno)
        if Q:
            self.HFILE(indent, '    Q_OBJECT', lineno)
            self.HFILE(indent, '', lineno)
        self.HFILE(indent, 'public:', lineno)
        self.namespace = name
        return True
        
    def C_CPPCLASS(self, indent, line, lineno, *args):
        return self.CPPCLASS(indent, line, lineno, *args, P=False)
        
    def QC_CPPCLASS(self, indent, line, lineno, *args):
        return self.CPPCLASS(indent, line, lineno, *args, P=False, Q=True)
        
    def QD_CPPCLASS(self, indent, line, lineno, *args):
        return self.CPPCLASS(indent, line, lineno, *args, D=True, Q=True)
        
    def CPPCLASS_END(self, indent, line, lineno, *args):
        self.namespace = None
        self.HFILE(indent, '};', lineno)
        self.HFILE(indent, '', lineno)
        return True
        
    def FUNC(self, indent, line, lineno, *args):
        def norm_type(t, types):
            t = t if type(t) is str else types.get(t.__name__, t.__name__)
            if t and t[-1].isalnum():
                t += ' '
            return t
        def norm_types(args, types):
            def gett(i, t):
                if type(t) is not int:
                    return t
                assert t == 1
                return gett(i+t, args[i+t][0])
            for i, (t, n) in enumerate(args):
                yield norm_type(gett(i, t), types), n
        argstr = lambda args: ', '.join((t+n) for t,n in args)
        if self.makrolineno == 0:
            self.makrodata = parse_mods(line)
        else:
            f_mod, m_mods, p_mods, c_mods = self.makrodata
            returntype, name, args = parse_pfunc(line)
            preturntype = norm_type(returntype, pyxrtypes)
            creturntype = norm_type(returntype, cpprtypes)
            if name.startswith('slot_'):
                creturntype = 'Q_INVOKABLE ' + creturntype
                name = name.split('_', 1)[1]
            elif name == '__init__':
                preturntype = ''
                creturntype = ''
                name = self.namespace
            if self.namespace is None and preturntype.split(' ')[0] not in ['cdef', 'cpdef']:
                preturntype = 'cdef ' + preturntype
            cname = name if self.namespace is None else self.namespace + '::' + name
            pyxargs = list(norm_types(args, pyxtypes))
            cppargs = list(norm_types(args, cpptypes))
            if f_mod == 'P':
                self.PYXFILE('{}{}({}){}:'.format(preturntype, name, argstr(pyxargs), p_mods))
            elif f_mod == 'H':
                self.HFILE(indent, '{}{}({}){}'.format(creturntype, name, argstr(cppargs), c_mods), lineno)
            elif f_mod == 'C':
                self.CFILE(indent, '{}{}({}){}'.format(creturntype, cname, argstr(cppargs), c_mods), lineno)
            else:
                assert False, f_mod
            if bremove(m_mods, 'H'):
                self.HFILE(indent, '{}{}({}){};'.format(creturntype, name, argstr(cppargs), c_mods), lineno)
            if bremove(m_mods, 'D'):
                self.PXDFILE(indent, '{}{}({}){}'.format(preturntype, name, argstr(pyxargs), p_mods), lineno)
            if bremove(m_mods, 'P'):
                self.PYXFILE('{}{}({}){}'.format(preturntype, name, argstr(pyxargs), p_mods))
            assert not m_mods, m_mods
            return True
            
    def Q_PROPERTY(self, indent, line, lineno, *args):
        if not line:
            return
        match = re.match(r"^prop_(\w+) = '(.*)'$", line.strip(), re.ASCII)
        pname = match.group(1)
        assert pname.isidentifier(), pname
        ptype, *flags = match.group(2).split()
        assert all(f.isupper() for f in flags), flags
        pargs = []
        def pop(flag):
            if flag in flags:
                flags.remove(flag)
                return True
            return False
        if pop('READ'):
            pargs += ['READ', 'get_' + pname]
        if pop('WRITE'):
            pargs += ['WRITE', 'set_' + pname]
        if pop('MEMBER'):
            fname = 'm_' + pname
            pargs += ['MEMBER', fname]
            self.HFILE(indent, '{} {};'.format(ptype, fname), lineno)
        if pop('NOTIFY'):
            fname = pname + '_changed'
            pargs += ['NOTIFY', fname]
            self.PXDFILE(indent, 'void {}() nogil'.format(fname), lineno)
            self.HFILE(indent, 'Q_SIGNAL void {}();'.format(fname), lineno)
        if flags:
            assert False, flags
        self.HFILE(indent, 'Q_PROPERTY ({} {} {})'.format(ptype, pname, ' '.join(pargs)), lineno)
        
    def Q_SIGNAL(self, indent, line, lineno, *args):
        if not line or line[0] == '#':
            return
        match = re.match(r"^(\w+) = pyqtSignal\((.*)\)$", line.strip(), re.ASCII)
        assert match is not None, line
        pname = match.group(1)
        assert pname.isidentifier(), pname
        types = ', '.join(cpptype(t.strip()) for t in match.group(2).split(','))
        self.HFILE(indent, 'Q_SIGNAL void {}({});'.format(pname, types), lineno)
        
    def ACTION(self, indent, line, lineno, *args):
        if not line:
            return
        match = re.match(r"^action_(\w+) = (\w+)$", line.strip(), re.ASCII)
        assert match is not None, line
        name = match.group(1)
        assert name.isidentifier(), name
        func = match.group(2)
        assert func.isidentifier(), func
        self.HFILE(indent, 'Q_INVOKABLE void on_action_{}_triggered();'.format(name), lineno)
        self.PYXFILE('void on_action_{0}_triggered() with gil:  {1}.app.on_action_{0}_triggered()'.format(name, func))
        
    



# generated from pybiklib/ext/gl_ogl.pxd

from libc.stdio cimport printf, puts
cimport gl_[[GLVARIANT]] as gl

print('Importing module:', __name__)
print('  GL-type: [[GLVARIANT]]')

def set_debug_flags(module):
    pass



cdef void _gl_check_error() nogil:
    cdef gl.GLenum err
    err = gl.glGetError()
    while err:
        if err == gl.GL_INVALID_ENUM:
            puts('glerror: GL_INVALID_ENUM')
        elif err == gl.GL_INVALID_VALUE:
            puts('glerror: GL_INVALID_VALUE')
        elif err == gl.GL_INVALID_OPERATION:
            puts('glerror: GL_INVALID_OPERATION')
        elif err == gl.GL_OUT_OF_MEMORY:
            puts('glerror: GL_OUT_OF_MEMORY')
        else:
            printf('gl unknown: %x\n', err)
        err = gl.glGetError()


# functions from /usr/include/GL/gl.h:

cdef void  glClearColor( GLclampf red, GLclampf green, GLclampf blue, GLclampf alpha ) nogil:
    puts('''gl:  void  glClearColor( GLclampf red, GLclampf green, GLclampf blue, GLclampf alpha )''')
    gl. glClearColor(red, green, blue, alpha)
    _gl_check_error()
cdef void  glClear( GLbitfield mask ) nogil:
    puts('''gl:  void  glClear( GLbitfield mask )''')
    gl. glClear(mask)
    _gl_check_error()
cdef void  glCullFace( GLenum mode ) nogil:
    puts('''gl:  void  glCullFace( GLenum mode )''')
    gl. glCullFace(mode)
    _gl_check_error()
cdef void  glFrontFace( GLenum mode ) nogil:
    puts('''gl:  void  glFrontFace( GLenum mode )''')
    gl. glFrontFace(mode)
    _gl_check_error()
cdef void  glEnable( GLenum cap ) nogil:
    puts('''gl:  void  glEnable( GLenum cap )''')
    gl. glEnable(cap)
    _gl_check_error()
cdef void  glDisable( GLenum cap ) nogil:
    puts('''gl:  void  glDisable( GLenum cap )''')
    gl. glDisable(cap)
    _gl_check_error()
cdef GLboolean  glIsEnabled( GLenum cap ) nogil:
    puts('''gl:  GLboolean  glIsEnabled( GLenum cap )''')
    cdef GLboolean result = gl. glIsEnabled(cap)
    _gl_check_error()
    return result
cdef void  glGetBooleanv( GLenum pname, GLboolean *params ) nogil:
    puts('''gl:  void  glGetBooleanv( GLenum pname, GLboolean *params )''')
    gl. glGetBooleanv(pname, params)
    _gl_check_error()
cdef void  glGetFloatv( GLenum pname, GLfloat *params ) nogil:
    puts('''gl:  void  glGetFloatv( GLenum pname, GLfloat *params )''')
    gl. glGetFloatv(pname, params)
    _gl_check_error()
cdef void  glGetIntegerv( GLenum pname, GLint *params ) nogil:
    puts('''gl:  void  glGetIntegerv( GLenum pname, GLint *params )''')
    gl. glGetIntegerv(pname, params)
    _gl_check_error()
cdef GLenum  glGetError() nogil:
    puts('''gl:  GLenum  glGetError( void )''')
    return gl. glGetError()
cdef GLubyte *  glGetString( GLenum name ) nogil:
    puts('''gl:  GLubyte *  glGetString( GLenum name )''')
    cdef GLubyte * result = <GLubyte *>gl. glGetString(name)
    _gl_check_error()
    return result
cdef void  glViewport( GLint x, GLint y,
                                    GLsizei width, GLsizei height ) nogil:
    puts('''gl:  void  glViewport( GLint x, GLint y,
                                    GLsizei width, GLsizei height )''')
    gl. glViewport(x, y, width, height)
    _gl_check_error()
cdef void  glDrawArrays( GLenum mode, GLint first, GLsizei count ) nogil:
    puts('''gl:  void  glDrawArrays( GLenum mode, GLint first, GLsizei count )''')
    gl. glDrawArrays(mode, first, count)
    _gl_check_error()
cdef void  glReadPixels( GLint x, GLint y,
                                    GLsizei width, GLsizei height,
                                    GLenum format, GLenum type,
                                    GLvoid *pixels ) nogil:
    puts('''gl:  void  glReadPixels( GLint x, GLint y,
                                    GLsizei width, GLsizei height,
                                    GLenum format, GLenum type,
                                    GLvoid *pixels )''')
    gl. glReadPixels(x, y, width, height, format, type, pixels)
    _gl_check_error()
cdef void  glTexImage2D( GLenum target, GLint level,
                                    GLint internalFormat,
                                    GLsizei width, GLsizei height,
                                    GLint border, GLenum format, GLenum type,
                                    GLvoid *pixels ) nogil:
    puts('''gl:  void  glTexImage2D( GLenum target, GLint level,
                                    GLint internalFormat,
                                    GLsizei width, GLsizei height,
                                    GLint border, GLenum format, GLenum type,
                                    GLvoid *pixels )''')
    gl. glTexImage2D(target, level, internalFormat, width, height, border, format, type, pixels)
    _gl_check_error()
cdef void  glTexSubImage2D( GLenum target, GLint level,
                                       GLint xoffset, GLint yoffset,
                                       GLsizei width, GLsizei height,
                                       GLenum format, GLenum type,
                                       GLvoid *pixels ) nogil:
    puts('''gl:  void  glTexSubImage2D( GLenum target, GLint level,
                                       GLint xoffset, GLint yoffset,
                                       GLsizei width, GLsizei height,
                                       GLenum format, GLenum type,
                                       GLvoid *pixels )''')
    gl. glTexSubImage2D(target, level, xoffset, yoffset, width, height, format, type, pixels)
    _gl_check_error()
cdef void  glActiveTexture( GLenum texture ) nogil:
    puts('''gl:  void  glActiveTexture( GLenum texture )''')
    gl. glActiveTexture(texture)
    _gl_check_error()


# functions from /usr/include/GL/glext.h:

cdef void  glBindBuffer (GLenum target, GLuint buffer) nogil:
    puts('''gl:  void  glBindBuffer (GLenum target, GLuint buffer)''')
    gl. glBindBuffer (target, buffer)
    _gl_check_error()
cdef void  glDeleteBuffers (GLsizei n, GLuint *buffers) nogil:
    puts('''gl:  void  glDeleteBuffers (GLsizei n, GLuint *buffers)''')
    gl. glDeleteBuffers (n, buffers)
    _gl_check_error()
cdef void  glGenBuffers (GLsizei n, GLuint *buffers) nogil:
    puts('''gl:  void  glGenBuffers (GLsizei n, GLuint *buffers)''')
    gl. glGenBuffers (n, buffers)
    _gl_check_error()
cdef void  glBufferData (GLenum target, GLsizeiptr size, void *data, GLenum usage) nogil:
    puts('''gl:  void  glBufferData (GLenum target, GLsizeiptr size, void *data, GLenum usage)''')
    gl. glBufferData (target, size, data, usage)
    _gl_check_error()
cdef void  glBufferSubData (GLenum target, GLintptr offset, GLsizeiptr size, void *data) nogil:
    puts('''gl:  void  glBufferSubData (GLenum target, GLintptr offset, GLsizeiptr size, void *data)''')
    gl. glBufferSubData (target, offset, size, data)
    _gl_check_error()
cdef void  glAttachShader (GLuint program, GLuint shader) nogil:
    puts('''gl:  void  glAttachShader (GLuint program, GLuint shader)''')
    gl. glAttachShader (program, shader)
    _gl_check_error()
cdef void  glBindAttribLocation (GLuint program, GLuint index, GLchar *name) nogil:
    puts('''gl:  void  glBindAttribLocation (GLuint program, GLuint index, GLchar *name)''')
    gl. glBindAttribLocation (program, index, name)
    _gl_check_error()
cdef void  glCompileShader (GLuint shader) nogil:
    puts('''gl:  void  glCompileShader (GLuint shader)''')
    gl. glCompileShader (shader)
    _gl_check_error()
cdef GLuint  glCreateProgram () nogil:
    puts('''gl:  GLuint  glCreateProgram (void)''')
    cdef GLuint result = gl. glCreateProgram ()
    _gl_check_error()
    return result
cdef GLuint  glCreateShader (GLenum type) nogil:
    puts('''gl:  GLuint  glCreateShader (GLenum type)''')
    cdef GLuint result = gl. glCreateShader (type)
    _gl_check_error()
    return result
cdef void  glDeleteProgram (GLuint program) nogil:
    puts('''gl:  void  glDeleteProgram (GLuint program)''')
    gl. glDeleteProgram (program)
    _gl_check_error()
cdef void  glDeleteShader (GLuint shader) nogil:
    puts('''gl:  void  glDeleteShader (GLuint shader)''')
    gl. glDeleteShader (shader)
    _gl_check_error()
cdef void  glDetachShader (GLuint program, GLuint shader) nogil:
    puts('''gl:  void  glDetachShader (GLuint program, GLuint shader)''')
    gl. glDetachShader (program, shader)
    _gl_check_error()
cdef void  glDisableVertexAttribArray (GLuint index) nogil:
    puts('''gl:  void  glDisableVertexAttribArray (GLuint index)''')
    gl. glDisableVertexAttribArray (index)
    _gl_check_error()
cdef void  glEnableVertexAttribArray (GLuint index) nogil:
    puts('''gl:  void  glEnableVertexAttribArray (GLuint index)''')
    gl. glEnableVertexAttribArray (index)
    _gl_check_error()
cdef void  glGetActiveAttrib (GLuint program, GLuint index, GLsizei bufSize, GLsizei *length, GLint *size, GLenum *type, GLchar *name) nogil:
    puts('''gl:  void  glGetActiveAttrib (GLuint program, GLuint index, GLsizei bufSize, GLsizei *length, GLint *size, GLenum *type, GLchar *name)''')
    gl. glGetActiveAttrib (program, index, bufSize, length, size, type, name)
    _gl_check_error()
cdef void  glGetActiveUniform (GLuint program, GLuint index, GLsizei bufSize, GLsizei *length, GLint *size, GLenum *type, GLchar *name) nogil:
    puts('''gl:  void  glGetActiveUniform (GLuint program, GLuint index, GLsizei bufSize, GLsizei *length, GLint *size, GLenum *type, GLchar *name)''')
    gl. glGetActiveUniform (program, index, bufSize, length, size, type, name)
    _gl_check_error()
cdef GLint  glGetAttribLocation (GLuint program, GLchar *name) nogil:
    puts('''gl:  GLint  glGetAttribLocation (GLuint program, GLchar *name)''')
    cdef GLint result = gl. glGetAttribLocation (program, name)
    _gl_check_error()
    return result
cdef void  glGetProgramiv (GLuint program, GLenum pname, GLint *params) nogil:
    puts('''gl:  void  glGetProgramiv (GLuint program, GLenum pname, GLint *params)''')
    gl. glGetProgramiv (program, pname, params)
    _gl_check_error()
cdef void  glGetProgramInfoLog (GLuint program, GLsizei bufSize, GLsizei *length, GLchar *infoLog) nogil:
    puts('''gl:  void  glGetProgramInfoLog (GLuint program, GLsizei bufSize, GLsizei *length, GLchar *infoLog)''')
    gl. glGetProgramInfoLog (program, bufSize, length, infoLog)
    _gl_check_error()
cdef void  glGetShaderiv (GLuint shader, GLenum pname, GLint *params) nogil:
    puts('''gl:  void  glGetShaderiv (GLuint shader, GLenum pname, GLint *params)''')
    gl. glGetShaderiv (shader, pname, params)
    _gl_check_error()
cdef void  glGetShaderInfoLog (GLuint shader, GLsizei bufSize, GLsizei *length, GLchar *infoLog) nogil:
    puts('''gl:  void  glGetShaderInfoLog (GLuint shader, GLsizei bufSize, GLsizei *length, GLchar *infoLog)''')
    gl. glGetShaderInfoLog (shader, bufSize, length, infoLog)
    _gl_check_error()
cdef GLint  glGetUniformLocation (GLuint program, GLchar *name) nogil:
    puts('''gl:  GLint  glGetUniformLocation (GLuint program, GLchar *name)''')
    cdef GLint result = gl. glGetUniformLocation (program, name)
    _gl_check_error()
    return result
cdef void  glLinkProgram (GLuint program) nogil:
    puts('''gl:  void  glLinkProgram (GLuint program)''')
    gl. glLinkProgram (program)
    _gl_check_error()
cdef void  glShaderSource (GLuint shader, GLsizei count, GLchar **string, GLint *length) nogil:
    puts('''gl:  void  glShaderSource (GLuint shader, GLsizei count, GLchar **string, GLint *length)''')
    gl. glShaderSource (shader, count, <const GLchar *const*>string, length)
    _gl_check_error()
cdef void  glUseProgram (GLuint program) nogil:
    puts('''gl:  void  glUseProgram (GLuint program)''')
    gl. glUseProgram (program)
    _gl_check_error()
cdef void  glUniform1i (GLint location, GLint v0) nogil:
    puts('''gl:  void  glUniform1i (GLint location, GLint v0)''')
    gl. glUniform1i (location, v0)
    _gl_check_error()
cdef void  glUniformMatrix4fv (GLint location, GLsizei count, GLboolean transpose, GLfloat *value) nogil:
    puts('''gl:  void  glUniformMatrix4fv (GLint location, GLsizei count, GLboolean transpose, GLfloat *value)''')
    gl. glUniformMatrix4fv (location, count, transpose, value)
    _gl_check_error()
cdef void  glValidateProgram (GLuint program) nogil:
    puts('''gl:  void  glValidateProgram (GLuint program)''')
    gl. glValidateProgram (program)
    _gl_check_error()
cdef void  glVertexAttribPointer (GLuint index, GLint size, GLenum type, GLboolean normalized, GLsizei stride, void *pointer) nogil:
    puts('''gl:  void  glVertexAttribPointer (GLuint index, GLint size, GLenum type, GLboolean normalized, GLsizei stride, void *pointer)''')
    gl. glVertexAttribPointer (index, size, type, normalized, stride, pointer)
    _gl_check_error()



#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2012-2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.



import sys, os
sys.path.insert(0, '.')
import re
from pathlib import Path
from textwrap import fill, wrap

from pybiklib import config

try:
    _
except NameError:
    import builtins
    builtins.__dict__['_'] = lambda s: s

footnote_first_idx = 1

def underline(title, underline_char='='):
    return '{}\n{}'.format(title, underline_char*len(title))

def clean_links(text):
    global footnote_first_idx
    footnote_cur_idx = 0
    footnotes = []
    def add_footnote(match):
        nonlocal footnote_cur_idx
        if match.group(2) != '|':
            return match.group(0)
        elif match.group(1) != '':
            footnote = match.group(1)
            if footnote in footnotes:
                footnote_cur_idx = footnotes.index(footnote) + footnote_first_idx
            else:
                footnote_cur_idx = len(footnotes) + footnote_first_idx
                footnotes.append(footnote)
            return ''
        else:
            return ' [{}]'.format(footnote_cur_idx)
    text = re.sub(r'<(.*?)(\|?)>', add_footnote, text)
    footnotes = ['[{}] <{}>'.format(i+footnote_first_idx, footnote) for i, footnote in enumerate(footnotes)]
    footnote_first_idx += len(footnotes)
    return text, footnotes
    
def wrap_text(text):
    for line in text.splitlines():
        if not line:
            yield ''
        elif len(line) <= 78:
            yield ' '+line if line.startswith(('* ', ' ')) else line
        elif line.startswith(('* ', ' ')):
            if ': ' in line:
                line1, line2 = line.split(': ', 1)
                assert len(line1) <= 78
                yield ' '+line1+':'
                yield from wrap(line2, width=78, initial_indent='   ', subsequent_indent='   ')
            else:
                yield from wrap(line, width=78, initial_indent=' ', subsequent_indent=' ')
        else:
            yield from wrap(line, width=78)
            
def lp_link(text):
    from subprocess import check_output
    fileid = check_output(('bzr', 'file-id', text), universal_newlines=True).strip()
    filename = os.path.basename(text)
    return '{}/head:/{}/{}'.format(config.REPOSITORY_DOWNLOAD, fileid, filename)
    
def xmlmarkup(lines):
    text = ''
    islist = False
    for line in lines:
        if line.startswith('* '):
            line = line[2:]
            assert line.strip() == line
            if islist:
                text += '</li>'
            else:
                text += '\n    <ul>'
            text += '\n      <li>'
            text += line
            islist = True
        elif line.startswith('  '):
            line = line[2:]
            assert line.strip() == line
            assert islist
            text += ' '+line
        else:
            assert line.strip() == line
            if islist:
                text += '</li>'
                text += '\n    </ul>'
            if line:
                text += '\n    <p>'+line+'</p>'
            islist = False
    if islist:
        text += '</li>'
        text += '\n    </ul>'
    text += '\n  '
    return text
    
def parse_NEWS():
    class S:    version, sep1, list1, listN, textN = range(5)
    nstate = S.version
    with open('NEWS', 'rt', encoding='utf-8') as f:
        text = f.read()
    news = []
    # news[i] : release i
    # news[i][0] : version of release i
    # news[i][1] : date of release i
    # news[i][2] : list of changes in release i
    # news[i][2][j] : change j in release i
    # news[i][2][j][0] : text of change j in release i
    # news[i][2][j][1] : sublist of change j in release i
    # news[i][2][j][1][k] : subitem k of change j in release i
    for line in text.splitlines():
        if nstate == S.version:
            assert line == line.strip()
            match = re.match(r'^(\w*) ([0-9.]*) \((\d\d\d\d-\d\d-\d\d|unreleased)\):$', line, flags=re.ASCII)
            name, version, date = match.groups()
            assert name == config.APPNAME
            if len(news) == 0:
                assert version == config.VERSION_SHORT
                assert date == config.RELEASE_DATE
            else:
                assert date != 'unreleased'
            news.append((version, date, []))
            nstate = S.sep1
        elif nstate == S.sep1:
            assert line == '', line
            nstate = S.list1
        elif nstate == S.list1:
            assert len(news[-1][2]) == 0
            if line.startswith('  * '):
                line = line[4:]
                assert line == line.strip()
                news[-1][2].append((line, []))
                nstate = S.listN
            else:
                assert line.startswith('  ')
                line = line[2:]
                assert line == line.strip()
                news[-1][2].append((line, []))
                nstate = S.textN
        elif nstate == S.listN:
            assert len(news[-1][2]) > 0
            if line.startswith('  * '):
                line = line[4:]
                assert line == line.strip()
                news[-1][2].append((line, []))
                nstate = S.listN
            elif line.startswith('    - '):
                line = line[6:]
                assert line == line.strip()
                news[-1][2][-1][1].append(line)
                nstate = S.listN
            elif line.startswith('    '):
                line = line.lstrip()
                assert line == line.strip()
                assert line and line[0] not in list(' *-')
                assert len(news[-1][2][-1][1]) == 0
                news[-1][2][-1] = (news[-1][2][-1][0] + ' ' + line, [])
                nstate = S.listN
            else:
                assert line == '', line
                nstate = S.version
        elif nstate == S.textN:
            assert len(news[-1][2]) == 1
            assert len(news[-1][2][0][1]) == 0
            if line.startswith('  '):
                line = line[2:]
                assert line == line.strip()
                assert line and line[0] not in list(' *-')
                news[-1][2][0] = (news[-1][2][0][0] + ' ' + line, [])
                nstate = S.listN
            elif line.startswith('    - '):
                line = line[6:]
                assert line == line.strip()
                news[-1][2][-1][1].append(line)
                nstate = S.listN
            else:
                assert line == '', line
                nstate = S.version
    assert all(len(n[2]) for n in news)
    return news
    
def xmlrelease(news):
    text = ''
    for version, date, changes in news:
        text += '\n    <release version="{}" date="{}">'.format(version, date)
        text += '\n      <description>'
        one = len(changes) == 1
        if not one:
            text += '\n        <ul>'
        for line, sublist in changes:
            text += '\n          '+('<p>' if one else '<li>')+line
            if sublist:
                text += '\n            <ul>'
                for line in sublist:
                    text += '\n              <li>'+line+'</li>'
                text += '\n            </ul>'
                text += '\n          '
            text += '</p>' if one else '</li>'
        if not one:
            text += '\n        </ul>'
        text += '\n      </description>'
        text += '\n    </release>'
    return text
    
            
class TemplateStr (str):
    __slots__ = ()
    linefmt = None
    delimiters = None
    footnotes = []
    mark = ''
    indent = None
    replacevals = None
    
    @classmethod
    def reset(cls):
        cls.linefmt = None
        cls.delimiters = None
        cls.footnotes.clear()
        cls.mark = ''
        
    def __format__(self, fmt):
        if fmt == 'line':
            text = self
            fmt = self.linefmt
            self.__class__.linefmt = None
        elif self == '#':
            self.__class__.linefmt = fmt
            return ''
        elif self == '##':
            self.__class__.linefmt = 'skip'
            return ''
        elif self == '#delimiters':
            delims = fmt.split()
            if delims:
                if len(delims) != 2:
                    raise IndexError('wrong count of delimiters: '+fmt)
                self.__class__.delimiters = re.escape(delims[0]) + r'(.*?)' + re.escape(delims[1]), r'{\1}'
            else:
                self.__class__.delimiters = None
            return ''
        elif self == '#footnotes':
            if not self.footnotes:
                return ''
            text = '\n' + '\n'.join(self.footnotes) + '\n'
            self.__class__.footnotes.clear()
        elif self == '#mark':
            self.__class__.mark = fmt
            return ''
        elif self == '#indent':
            self.__class__.indent = fmt
            return ''
        elif self == '#replace':
            self.__class__.replacevals = fmt.split(None, 1)
            return ''
        elif self == '#data':
            text, fmt = fmt.split(None, 1)
        elif self == '#news':
            #FIXME: should fail here, if NEWS not in depends
            text = parse_NEWS()
            text = xmlrelease(text)
        else:
            text = getattr(config, self)
            if callable(text):
                text = text()
        if not fmt:
            return text
        for s in fmt.split('.'):
            if s == 'wrap':
                text = '\n'.join(wrap_text(text))
            elif s == 'deb_text':
                text = '\n '.join(l or '.' for l in text.splitlines())
            elif s == 'links':
                text, footnotes = clean_links(text)
                self.__class__.footnotes += footnotes
            elif s == 'join':
                if isinstance(text, (tuple, list)):
                    text = '\n'.join(text)
                else:
                    text = text.replace('\n\n', '  ').replace('\n', ' ')
            elif s == 'indent':
                indent = self.indent or ' '
                text = '\n'.join((l and indent+l) for l in text.split('\n'))
            elif s == 'replace':
                rsrc, rdst = self.replacevals
                text = text.replace(rsrc, rdst)
            elif s == 'skip':
                text = ''
            elif s == 'xmlmarkup':
                text = xmlmarkup(text)
            elif s == 'lp-link':
                text = lp_link(text)
            else:
                raise ValueError('unknown format: ' + fmt)
        return text
        
        
class TemplateKeyDict (dict):
    __slots__ = ()
    def __getitem__(self, key):
        if key and key[0] in ["'", '"']:
            import ast
            return TemplateStr(ast.literal_eval(key))
        return TemplateStr(key)
        
        
def get_template_filename(templatename):
    if templatename == os.path.basename(templatename):
        templatename = os.path.join(config.appdata_dir, 'templates', templatename)
    templatename += '.in'
    return templatename
    
def create_doc(templatename, filename=None, *, skip=None):
    global footnote_first_idx
    footnote_first_idx = 1
    if not filename:
        filename = templatename
    templatename = get_template_filename(templatename)
    with open(templatename, 'rt', encoding='utf-8') as f:
        template = f.readlines()
    text = ''
    prev_linelen = None
    skipstate = 0
    TemplateStr.reset()
    for lineno, line in enumerate(template):
        line = line.rstrip()
        if skip == line:
            skipstate = 1
            continue
        elif skipstate == 3:
            skipstate = 0
        elif skipstate:
            skipstate = 1 if line else skipstate+1
            continue
        if prev_linelen is not None and 3 <= len(line) == line.count(line[0]):
            line = line[0] * prev_linelen
        else:
            if TemplateStr.delimiters is not None:
                line = line.replace('{','{{').replace('}','}}')
                pattern, replace = TemplateStr.delimiters
                line = re.sub(pattern, replace, line)
            try:
                line = line.format_map(TemplateKeyDict())
                line = format(TemplateStr(line), 'line')
            except Exception as e:
                print('{}:{}: {}'.format(templatename, lineno+1, e))
                print('   ', line, end='')
                raise
            prev_linelen = len(line)
        if skip == TemplateStr.mark:
            continue
        text += line + '\n'
    with open(filename, 'wt', encoding='utf-8') as f:
        f.write(text)
    
    
def main():
    skip = None
    args = sys.argv[1:]
    if not args:
        print('usage: {} [--skip=PARAGRAPH] template[=filename]'.format(os.path.basename(sys.argv[0])))
        sys.exit(1)
    for arg in args:
        if arg.startswith('--skip='):
            skip = arg.split('=', 1)[1]
        elif not arg.startswith('--'):
            fileinfo = arg.split('=', 1)
            create_doc(*fileinfo, skip=skip)
        else:
            print('Unknown Option:', arg)
            sys.exit(1)
            
    
if __name__ == '__main__':
    try:
        _
    except NameError:
        __builtins__._ = lambda text: text
    main()
    




# generated from pybiklib/ext/gl_es2.pxd

from libc.stdio cimport printf, puts
cimport gl_[[GLVARIANT]] as gl

print('Importing module:', __name__)
print('  GL-type: [[GLVARIANT]]')

def set_debug_flags(module):
    pass



cdef void _gl_check_error() nogil:
    cdef gl.GLenum err
    err = gl.glGetError()
    while err:
        if err == gl.GL_INVALID_ENUM:
            puts('glerror: GL_INVALID_ENUM')
        elif err == gl.GL_INVALID_VALUE:
            puts('glerror: GL_INVALID_VALUE')
        elif err == gl.GL_INVALID_OPERATION:
            puts('glerror: GL_INVALID_OPERATION')
        elif err == gl.GL_OUT_OF_MEMORY:
            puts('glerror: GL_OUT_OF_MEMORY')
        else:
            printf('gl unknown: %x\n', err)
        err = gl.glGetError()


# functions from /usr/include/GLES2/gl2.h:

cdef void  glActiveTexture (GLenum texture) nogil:
    puts('''gl:  void  glActiveTexture (GLenum texture)''')
    gl. glActiveTexture (texture)
    _gl_check_error()
cdef void  glAttachShader (GLuint program, GLuint shader) nogil:
    puts('''gl:  void  glAttachShader (GLuint program, GLuint shader)''')
    gl. glAttachShader (program, shader)
    _gl_check_error()
cdef void  glBindAttribLocation (GLuint program, GLuint index, GLchar *name) nogil:
    puts('''gl:  void  glBindAttribLocation (GLuint program, GLuint index, GLchar *name)''')
    gl. glBindAttribLocation (program, index, name)
    _gl_check_error()
cdef void  glBindBuffer (GLenum target, GLuint buffer) nogil:
    puts('''gl:  void  glBindBuffer (GLenum target, GLuint buffer)''')
    gl. glBindBuffer (target, buffer)
    _gl_check_error()
cdef void  glBufferData (GLenum target, GLsizeiptr size, void *data, GLenum usage) nogil:
    puts('''gl:  void  glBufferData (GLenum target, GLsizeiptr size, void *data, GLenum usage)''')
    gl. glBufferData (target, size, data, usage)
    _gl_check_error()
cdef void  glBufferSubData (GLenum target, GLintptr offset, GLsizeiptr size, void *data) nogil:
    puts('''gl:  void  glBufferSubData (GLenum target, GLintptr offset, GLsizeiptr size, void *data)''')
    gl. glBufferSubData (target, offset, size, data)
    _gl_check_error()
cdef void  glClear (GLbitfield mask) nogil:
    puts('''gl:  void  glClear (GLbitfield mask)''')
    gl. glClear (mask)
    _gl_check_error()
cdef void  glClearColor (GLfloat red, GLfloat green, GLfloat blue, GLfloat alpha) nogil:
    puts('''gl:  void  glClearColor (GLfloat red, GLfloat green, GLfloat blue, GLfloat alpha)''')
    gl. glClearColor (red, green, blue, alpha)
    _gl_check_error()
cdef void  glCompileShader (GLuint shader) nogil:
    puts('''gl:  void  glCompileShader (GLuint shader)''')
    gl. glCompileShader (shader)
    _gl_check_error()
cdef GLuint  glCreateProgram () nogil:
    puts('''gl:  GLuint  glCreateProgram (void)''')
    cdef GLuint result = gl. glCreateProgram ()
    _gl_check_error()
    return result
cdef GLuint  glCreateShader (GLenum type) nogil:
    puts('''gl:  GLuint  glCreateShader (GLenum type)''')
    cdef GLuint result = gl. glCreateShader (type)
    _gl_check_error()
    return result
cdef void  glCullFace (GLenum mode) nogil:
    puts('''gl:  void  glCullFace (GLenum mode)''')
    gl. glCullFace (mode)
    _gl_check_error()
cdef void  glDeleteBuffers (GLsizei n, GLuint *buffers) nogil:
    puts('''gl:  void  glDeleteBuffers (GLsizei n, GLuint *buffers)''')
    gl. glDeleteBuffers (n, buffers)
    _gl_check_error()
cdef void  glDeleteProgram (GLuint program) nogil:
    puts('''gl:  void  glDeleteProgram (GLuint program)''')
    gl. glDeleteProgram (program)
    _gl_check_error()
cdef void  glDeleteShader (GLuint shader) nogil:
    puts('''gl:  void  glDeleteShader (GLuint shader)''')
    gl. glDeleteShader (shader)
    _gl_check_error()
cdef void  glDetachShader (GLuint program, GLuint shader) nogil:
    puts('''gl:  void  glDetachShader (GLuint program, GLuint shader)''')
    gl. glDetachShader (program, shader)
    _gl_check_error()
cdef void  glDisable (GLenum cap) nogil:
    puts('''gl:  void  glDisable (GLenum cap)''')
    gl. glDisable (cap)
    _gl_check_error()
cdef void  glDisableVertexAttribArray (GLuint index) nogil:
    puts('''gl:  void  glDisableVertexAttribArray (GLuint index)''')
    gl. glDisableVertexAttribArray (index)
    _gl_check_error()
cdef void  glDrawArrays (GLenum mode, GLint first, GLsizei count) nogil:
    puts('''gl:  void  glDrawArrays (GLenum mode, GLint first, GLsizei count)''')
    gl. glDrawArrays (mode, first, count)
    _gl_check_error()
cdef void  glEnable (GLenum cap) nogil:
    puts('''gl:  void  glEnable (GLenum cap)''')
    gl. glEnable (cap)
    _gl_check_error()
cdef void  glEnableVertexAttribArray (GLuint index) nogil:
    puts('''gl:  void  glEnableVertexAttribArray (GLuint index)''')
    gl. glEnableVertexAttribArray (index)
    _gl_check_error()
cdef void  glFrontFace (GLenum mode) nogil:
    puts('''gl:  void  glFrontFace (GLenum mode)''')
    gl. glFrontFace (mode)
    _gl_check_error()
cdef void  glGenBuffers (GLsizei n, GLuint *buffers) nogil:
    puts('''gl:  void  glGenBuffers (GLsizei n, GLuint *buffers)''')
    gl. glGenBuffers (n, buffers)
    _gl_check_error()
cdef void  glGetActiveAttrib (GLuint program, GLuint index, GLsizei bufSize, GLsizei *length, GLint *size, GLenum *type, GLchar *name) nogil:
    puts('''gl:  void  glGetActiveAttrib (GLuint program, GLuint index, GLsizei bufSize, GLsizei *length, GLint *size, GLenum *type, GLchar *name)''')
    gl. glGetActiveAttrib (program, index, bufSize, length, size, type, name)
    _gl_check_error()
cdef void  glGetActiveUniform (GLuint program, GLuint index, GLsizei bufSize, GLsizei *length, GLint *size, GLenum *type, GLchar *name) nogil:
    puts('''gl:  void  glGetActiveUniform (GLuint program, GLuint index, GLsizei bufSize, GLsizei *length, GLint *size, GLenum *type, GLchar *name)''')
    gl. glGetActiveUniform (program, index, bufSize, length, size, type, name)
    _gl_check_error()
cdef GLint  glGetAttribLocation (GLuint program, GLchar *name) nogil:
    puts('''gl:  GLint  glGetAttribLocation (GLuint program, GLchar *name)''')
    cdef GLint result = gl. glGetAttribLocation (program, name)
    _gl_check_error()
    return result
cdef void  glGetBooleanv (GLenum pname, GLboolean *data) nogil:
    puts('''gl:  void  glGetBooleanv (GLenum pname, GLboolean *data)''')
    gl. glGetBooleanv (pname, data)
    _gl_check_error()
cdef GLenum  glGetError () nogil:
    puts('''gl:  GLenum  glGetError (void)''')
    return gl. glGetError ()
cdef void  glGetFloatv (GLenum pname, GLfloat *data) nogil:
    puts('''gl:  void  glGetFloatv (GLenum pname, GLfloat *data)''')
    gl. glGetFloatv (pname, data)
    _gl_check_error()
cdef void  glGetIntegerv (GLenum pname, GLint *data) nogil:
    puts('''gl:  void  glGetIntegerv (GLenum pname, GLint *data)''')
    gl. glGetIntegerv (pname, data)
    _gl_check_error()
cdef void  glGetProgramiv (GLuint program, GLenum pname, GLint *params) nogil:
    puts('''gl:  void  glGetProgramiv (GLuint program, GLenum pname, GLint *params)''')
    gl. glGetProgramiv (program, pname, params)
    _gl_check_error()
cdef void  glGetProgramInfoLog (GLuint program, GLsizei bufSize, GLsizei *length, GLchar *infoLog) nogil:
    puts('''gl:  void  glGetProgramInfoLog (GLuint program, GLsizei bufSize, GLsizei *length, GLchar *infoLog)''')
    gl. glGetProgramInfoLog (program, bufSize, length, infoLog)
    _gl_check_error()
cdef void  glGetShaderiv (GLuint shader, GLenum pname, GLint *params) nogil:
    puts('''gl:  void  glGetShaderiv (GLuint shader, GLenum pname, GLint *params)''')
    gl. glGetShaderiv (shader, pname, params)
    _gl_check_error()
cdef void  glGetShaderInfoLog (GLuint shader, GLsizei bufSize, GLsizei *length, GLchar *infoLog) nogil:
    puts('''gl:  void  glGetShaderInfoLog (GLuint shader, GLsizei bufSize, GLsizei *length, GLchar *infoLog)''')
    gl. glGetShaderInfoLog (shader, bufSize, length, infoLog)
    _gl_check_error()
cdef GLubyte * glGetString (GLenum name) nogil:
    puts('''gl:  GLubyte * glGetString (GLenum name)''')
    cdef GLubyte *result = <GLubyte *>gl. glGetString (name)
    _gl_check_error()
    return result
cdef GLint  glGetUniformLocation (GLuint program, GLchar *name) nogil:
    puts('''gl:  GLint  glGetUniformLocation (GLuint program, GLchar *name)''')
    cdef GLint result = gl. glGetUniformLocation (program, name)
    _gl_check_error()
    return result
cdef GLboolean  glIsEnabled (GLenum cap) nogil:
    puts('''gl:  GLboolean  glIsEnabled (GLenum cap)''')
    cdef GLboolean result = gl. glIsEnabled (cap)
    _gl_check_error()
    return result
cdef void  glLinkProgram (GLuint program) nogil:
    puts('''gl:  void  glLinkProgram (GLuint program)''')
    gl. glLinkProgram (program)
    _gl_check_error()
cdef void  glReadPixels (GLint x, GLint y, GLsizei width, GLsizei height, GLenum format, GLenum type, void *pixels) nogil:
    puts('''gl:  void  glReadPixels (GLint x, GLint y, GLsizei width, GLsizei height, GLenum format, GLenum type, void *pixels)''')
    gl. glReadPixels (x, y, width, height, format, type, pixels)
    _gl_check_error()
cdef void  glShaderSource (GLuint shader, GLsizei count, GLchar **string, GLint *length) nogil:
    puts('''gl:  void  glShaderSource (GLuint shader, GLsizei count, GLchar **string, GLint *length)''')
    gl. glShaderSource (shader, count, <const GLchar *const*>string, length)
    _gl_check_error()
cdef void  glTexImage2D (GLenum target, GLint level, GLint internalformat, GLsizei width, GLsizei height, GLint border, GLenum format, GLenum type, void *pixels) nogil:
    puts('''gl:  void  glTexImage2D (GLenum target, GLint level, GLint internalformat, GLsizei width, GLsizei height, GLint border, GLenum format, GLenum type, void *pixels)''')
    gl. glTexImage2D (target, level, internalformat, width, height, border, format, type, pixels)
    _gl_check_error()
cdef void  glTexSubImage2D (GLenum target, GLint level, GLint xoffset, GLint yoffset, GLsizei width, GLsizei height, GLenum format, GLenum type, void *pixels) nogil:
    puts('''gl:  void  glTexSubImage2D (GLenum target, GLint level, GLint xoffset, GLint yoffset, GLsizei width, GLsizei height, GLenum format, GLenum type, void *pixels)''')
    gl. glTexSubImage2D (target, level, xoffset, yoffset, width, height, format, type, pixels)
    _gl_check_error()
cdef void  glUniform1i (GLint location, GLint v0) nogil:
    puts('''gl:  void  glUniform1i (GLint location, GLint v0)''')
    gl. glUniform1i (location, v0)
    _gl_check_error()
cdef void  glUniformMatrix4fv (GLint location, GLsizei count, GLboolean transpose, GLfloat *value) nogil:
    puts('''gl:  void  glUniformMatrix4fv (GLint location, GLsizei count, GLboolean transpose, GLfloat *value)''')
    gl. glUniformMatrix4fv (location, count, transpose, value)
    _gl_check_error()
cdef void  glUseProgram (GLuint program) nogil:
    puts('''gl:  void  glUseProgram (GLuint program)''')
    gl. glUseProgram (program)
    _gl_check_error()
cdef void  glValidateProgram (GLuint program) nogil:
    puts('''gl:  void  glValidateProgram (GLuint program)''')
    gl. glValidateProgram (program)
    _gl_check_error()
cdef void  glVertexAttribPointer (GLuint index, GLint size, GLenum type, GLboolean normalized, GLsizei stride, void *pointer) nogil:
    puts('''gl:  void  glVertexAttribPointer (GLuint index, GLint size, GLenum type, GLboolean normalized, GLsizei stride, void *pointer)''')
    gl. glVertexAttribPointer (index, size, type, normalized, stride, pointer)
    _gl_check_error()
cdef void  glViewport (GLint x, GLint y, GLsizei width, GLsizei height) nogil:
    puts('''gl:  void  glViewport (GLint x, GLint y, GLsizei width, GLsizei height)''')
    gl. glViewport (x, y, width, height)
    _gl_check_error()


# functions from /usr/include/GLES2/gl2ext.h:




#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2016-2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

try:
    from PyQt5.QtCore import (pyqtProperty, pyqtSignal, pyqtSlot, Qt, QObject, QVariant, QUrl, QStandardPaths,
                            QElapsedTimer, QFileSystemWatcher, QTimer, QByteArray,
                            QLocale, QTranslator, QLibraryInfo, QMetaObject, QPointF)
    from PyQt5.QtGui import (QColor, QImage, QIcon, QKeySequence, QTextDocumentFragment, QSurfaceFormat,
                            QOpenGLTexture, QOpenGLFramebufferObject, QTransform, QPixmap, QCursor,
                            QOpenGLContext, QOpenGLDebugLogger, QIconEngine)
    from PyQt5.QtQml import QQmlEngine, QQmlComponent
    from PyQt5.QtWidgets import QApplication, QMainWindow, QLineEdit, QLabel, QOpenGLWidget, QDialog, QStyledItemDelegate
    from PyQt5.QtQuick import QQuickView, QQuickItem
except ImportError as e:
    print('Need PyQt5 in pure python mode:', e)
    raise SystemExit(1)
    
import OpenGL.GL as gl

NULL = None
Format_RGBA8888 = QImage.Format_RGBA8888
PortableText = QKeySequence.PortableText
MoveToNextWord = QKeySequence.MoveToNextWord
MoveToPreviousWord = QKeySequence.MoveToPreviousWord
fromString = QKeySequence.fromString
standardLocations = QStandardPaths.standardLocations
PicturesLocation = QStandardPaths.PicturesLocation
HomeLocation = QStandardPaths.HomeLocation
fromPlainText = QTextDocumentFragment.fromPlainText
setDefaultFormat = QSurfaceFormat.setDefaultFormat
defaultFormat = QSurfaceFormat.defaultFormat
Target2D = QOpenGLTexture.Target2D
RGBA32F = QOpenGLTexture.RGBA32F
Linear = QOpenGLTexture.Linear
Depth = QOpenGLFramebufferObject.Depth
fromLocalFile = QUrl.fromLocalFile
Asynchronous = QQmlComponent.Asynchronous
fromImage = QPixmap.fromImage
location = QLibraryInfo.location
TranslationsPath = QLibraryInfo.TranslationsPath
QLocale_system = QLocale.system
arguments = QApplication.arguments
setAttribute = QApplication.setAttribute
OpenGLES = QSurfaceFormat.OpenGLES
DebugContext = QSurfaceFormat.DebugContext
openGLModuleType = QOpenGLContext.openGLModuleType
LibGL = QOpenGLContext.LibGL
LibGLES = QOpenGLContext.LibGLES
currentContext = QOpenGLContext.currentContext
SynchronousLogging = QOpenGLDebugLogger.SynchronousLogging
invokeMethod = QMetaObject.invokeMethod
SizeRootObjectToView = QQuickView.SizeRootObjectToView
GL_TEXTURE_2D = gl.GL_TEXTURE_2D


class MetaSlot(type(QObject)):
    def __new__(cls, name, bases, namespace):
        import inspect
        qtypemap = {'QString': str, 'QList<QObject*>': QVariant, 'QVariantList': QVariant}
        def member(proptype, propname, notify):
            def getter(self):
                try:
                    returnv = getattr(self, propname)
                except AttributeError:
                    returnv = proptype()
                return returnv
            def setter(self, v):
                setattr(self, propname, v)
                notify and notify.__get__(self).emit()
            return getter, setter
            
        for k, v in sorted(namespace.items()):
            if k.startswith('slot_'):
                spec = inspect.getfullargspec(v)
                argtypes = [spec.annotations[a] for a in spec.args[1:]]
                rtype = spec.annotations.get('return')
                fname = k.split('_', 1)[1]
                if rtype is None:
                    slot = pyqtSlot(*argtypes, name=fname)
                else:
                    slot = pyqtSlot(*argtypes, name=fname, result=rtype)
                assert fname not in namespace
                namespace[fname] = slot(v)
                del namespace[k]
            elif k.startswith('action_'):
                action = k.split('_', 1)[1]
                fname = 'on_action_%s_triggered' % action
                slot = pyqtSlot(name=fname)
                assert fname not in namespace
                def _mkactionfunc(action, pydata):
                    def on_action_X_triggered(self):
                        getattr(pydata.app, 'on_action_%s_triggered' % action)()
                    return on_action_X_triggered
                namespace[fname] = slot(_mkactionfunc(action, v))
                del namespace[k]
            elif k.startswith('prop_'):
                pname = k.split('_', 1)[1]
                ptype, *flags = v.split()
                try:
                    ptype = qtypemap.get(ptype) or eval(ptype.rstrip('*'))
                except NameError:
                    print('{}: name={}, type={!r}, flags={!r}'.format(k, pname, ptype, flags))
                    raise
                assert type(ptype) is not str or ptype.isalpha(), ptype
                pargs = {}
                getter = setter = notify = None
                def pop(flag):
                    if flag in flags:
                        flags.remove(flag)
                        return True
                    return False
                if pop('NOTIFY'):
                    fname = pname + '_changed'
                    assert fname not in namespace, fname
                    notify = pyqtSignal()
                    pargs['notify'] = notify
                    namespace[fname] = notify
                if pop('MEMBER'):
                    fname = 'm_' + pname
                    assert fname not in namespace, fname
                    getter, setter = member(ptype, fname, notify)
                    pargs['fget'], pargs['fset'] = getter, setter
                if pop('READ'):
                    fname = 'get_' + pname
                    assert fname in namespace, fname
                    getter = namespace[fname]
                    pargs['fget'] = getter
                if pop('WRITE'):
                    fname = 'set_' + pname
                    assert fname in namespace, fname
                    setter = namespace[fname]
                    pargs['fset'] = setter
                if flags:
                    assert False, flags
                assert pname not in namespace, pname
                namespace[pname] = pyqtProperty(ptype, **pargs)
                del namespace[k]
        return type(QObject)(name, bases, namespace)
        
    




#-*- coding:utf-8 -*-

#  Copyright © 2009-2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import pickle
from contextlib import suppress

from .model import empty_model
from . import cubestate
from .moves import MoveQueue


class GameState:
    def __init__(self):
        self.mark_before = False
        self.initial_state = cubestate.CubeState(empty_model)
        self.initial_state.set_solved()
        self.current_state = self.initial_state.copy()
        self.move_sequence = MoveQueue()
        
    def copy(self):
        game = GameState()
        game.mark_before = self.mark_before
        game.initial_state = self.initial_state.copy()
        game.current_state = self.current_state.copy()
        game.move_sequence = self.move_sequence.copy()
        return game
        
    def set_state(self, model, blocks, moves, position):
        self.initial_state = cubestate.CubeState(model)
        try:
            self.initial_state.parse_block(blocks)
        except ValueError as e:
            print('Invalid block list:', *e.args)
            print('  start with solved game')
            self.initial_state.set_solved()
        self.current_state = self.initial_state.copy()
        self.move_sequence = MoveQueue()
        if moves.startswith('native:'):
            moves = moves.split(':', 1)[1].lstrip()
        else:
            moves = ''
        pos, unused_cpos = self.move_sequence.parse(moves, position, model)
        self.goto_next_pos(pos)
        
    def get_state(self):
        model = self.initial_state.model
        blocks = self.initial_state.format_block()
        moves, position = self.move_sequence.format(model)
        return model, blocks, 'native: ' + moves, position
        
    @staticmethod
    def _load_game_file(filename, mtype):
        data = None
        if filename is not None:
            try:
                with suppress(FileNotFoundError):
                    with open(filename, 'rb') as datafile:
                        data = pickle.load(datafile)
            except Exception as e:
                print('error while loading game {0}: {1.__class__.__name__}: {1}'.format(mtype, e))
        if type(data) is not dict:
            data = {}
            
        data_games = None
        with suppress(KeyError):
            data_games = data['games']
        if type(data_games) is not dict:
            data_games = {}
            data['games'] = data_games
            
        data_type = None
        with suppress(KeyError):
            data_type = data_games[mtype]
        if type(data_type) is not dict:
            data_type = {}
            data_games[mtype] = data_type
        return data, data_type
        
    def load(self, filename, model):
        unused_data, data_type = self._load_game_file(filename, model.type)
        blocks, moves, position = data_type.get(model.size, ('solved', '', 0))
        self.set_state(model, blocks, moves, position)
        
    def save(self, filename):
        if filename is None:
            return
        model, blocks, moves, position = self.get_state()
        if model is empty_model:
            return
        basename = os.path.basename(filename)
        dirname = os.path.dirname(filename)
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname)
        data, data_type = self._load_game_file(filename, model.type)
        data_type[model.size] = blocks, moves, position
        import tempfile
        try:
            with tempfile.NamedTemporaryFile('wb', prefix=basename+'-temp-', dir=dirname, delete=False) as datafile:
                pickle.dump(data, datafile, protocol=4)
            os.rename(datafile.name, filename)
        except Exception as e:
            print('error while saving game {0}: {1.__class__.__name__}: {1}'.format(model.type, e))
            
    @classmethod
    def migrate_1to2(self, filename, mtype, msize, blocks, moves, position):
        data, data_type = self._load_game_file(filename, mtype)
        data_type[msize] = blocks, moves, position
        try:
            with open(filename, 'wb') as datafile:
                pickle.dump(data, datafile, protocol=4)
        except Exception as e:
            print('error during migrantion of {0}: {1.__class__.__name__}: {1}'.format(mtype, e))
            
    def random(self, count=-1):
        self.initial_state.set_solved()
        self.initial_state.random(count)
        self.current_state = self.initial_state.copy()
        self.move_sequence.reset()
        
    ### Funktions to control the cube for use in callbacks and plugins
    
    def step_back(self):
        '''One step back in the sequence of moves'''
        if self.move_sequence.at_start():
            return None
        self.mark_before = self.move_sequence.is_mark_current()
        self.move_sequence.retard()
        move = self.move_sequence.current().inverted()
        self.current_state.rotate_slice(move)
        return move
        
    def step_next(self):
        '''One step forward in the sequence of moves'''
        move = self.move_sequence.current()
        if move:
            self.mark_before = self.move_sequence.is_mark_current()
            self.current_state.rotate_slice(move)
            self.move_sequence.advance()
        return move
        
    def set_next(self, move):
        '''Make one new move.'''
        self.move_sequence.push_current(move)
        
    def goto_prev_mark(self):
        if self.move_sequence.at_start():
            return False
        while True:
            self.move_sequence.retard()
            move = self.move_sequence.current().inverted()
            self.current_state.rotate_slice(move)
            if self.move_sequence.is_mark_current():
                break
        return True
        
    def goto_start(self):
        self.current_state = self.initial_state.copy()
        self.move_sequence.rewind_start()
            
    def goto_next_mark(self):
        move = self.move_sequence.current()
        while move:
            self.current_state.rotate_slice(move)
            self.move_sequence.advance()
            if self.move_sequence.is_mark_current():
                break
            move = self.move_sequence.current()
        
    def goto_next_pos(self, pos):
        move = self.move_sequence.current()
        while move:
            if pos == self.move_sequence.current_place:
                break
            self.current_state.rotate_slice(move)
            self.move_sequence.advance()
            move = self.move_sequence.current()
        
    def set_code(self, code, pos):
        self.current_state = self.initial_state.copy()
        self.move_sequence.reset()
        pos, unused_cpos = self.move_sequence.parse(code, pos, self.current_state.model)
        self.goto_next_pos(pos)
        
    def add_moves(self, moves):
        for move in moves:
            self.current_state.rotate_slice(move)
            self.move_sequence.push(move)
            
    def add_code(self, code):
        for res in self.move_sequence.parse_iter(code, len(code), self.current_state.model):
            self.current_state.rotate_slice(res[0])
            
    def set_plugin_mode(self, plugin_mode):
        ''' Set plugin mode
            append:     append moves from the plugin to the game
            replace:    replace initial state and moves
            challenge:  replace initial state and clear moves, set the unsolved flag
        '''
        self.plugin_mode = plugin_mode
        
    def recalc_current_state(self):
        self.current_state = self.initial_state.copy()
        pos = self.move_sequence.current_place
        self.move_sequence.rewind_start()
        self.goto_next_pos(pos)
        
    # game transformations
        
    def set_as_initial_state(self):
        self.initial_state = self.current_state.copy()
        self.move_sequence.truncate_before()
        
    # cube editing
    
    def swap_block(self, blockpos, maxis, mslice, mdir):
        self.initial_state.swap_block(blockpos, maxis, mslice, mdir)
        self.goto_start()
        
    def rotate_block(self, blockpos, rdir):
        self.initial_state.rotate_block(blockpos, rdir)
        self.goto_start()
        




# -*- coding: utf-8 -*-

#  Copyright © 2013-2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, os
from glob import glob
from contextlib import suppress


def update_linguas():
    import polib
    min_percent_translated = 20
    cnt_skipped = 0
    with open('po/LINGUAS', 'wt', encoding='utf-8') as linguas:
        po_files = sorted(glob('po/*.po'))
        for po_file in po_files:
            percent_translated = polib.pofile(po_file).percent_translated()
            if percent_translated < min_percent_translated:
                print('skipping po file {!r} ({}%<{}% translated)'.format(
                            po_file, percent_translated, min_percent_translated),
                    file=sys.stderr)
                cnt_skipped += 1
            else:
                print(os.path.splitext(os.path.basename(po_file))[0], file=linguas)
        print('skipped {}/{} po files'.format(cnt_skipped, len(po_files)), file=sys.stderr)
                
def get_linguas():
    with open('po/LINGUAS', 'rt', encoding='utf-8') as linguas:
        return linguas.read().splitlines()
        
def generate_lang_translations():
    from pybiklib.dialogs import get_langname_from_code_func
    langname_from_code = get_langname_from_code_func()
    names = ((code, langname_from_code(code)) for code in get_linguas())
    names = {code: name for code, name in names if name is not None}
    print(names)
    
def generate_all_lang_translations(file=sys.stdout):
    import icu  # to be sure get_langname_from_code_func does not return a fallback function
    from subprocess import Popen, PIPE
    from ast import literal_eval
    from pprint import pformat
    
    processes = []
    env = os.environ.copy()
    for lang in get_linguas():
        env['LC_MESSAGES'] = lang+'.UTF-8'
        script = 'import buildlib.utils as m; m.generate_lang_translations()'
        print(lang, end=' ', flush=True)
        processes.append((lang, Popen(['python3', '-c', script], env=env, stdout=PIPE)))
    print()
    
    langs = {}
    for lang, process in processes:
        print(lang, end=' ', flush=True)
        outstr, unused_err = process.communicate()
        outstr = outstr.strip().decode('utf-8')
        try:
            langs[lang] = literal_eval(outstr)
        except SyntaxError:
            print('code for lang %r:' % lang, repr(outstr))
            raise
    print()
    
    print('# -*- coding: utf-8 -*-', file=file)
    print('# generated file\n', file=file)
    print('languages = (', file=file)
    print(pformat(langs), file=file)
    print(')\n', file=file)
    
def modify_file(filename, subs, count=1):
    from shutil import copymode
    import re
    print('modifying', filename)
    with open(filename, 'rt', encoding='utf-8') as f:
        rtext = f.read()
    wtext = rtext
    for pattern, repl in subs:
        wtext = re.sub(pattern, repl, wtext, count=count, flags=re.MULTILINE)
    if wtext == rtext:
        return False
    tmpfilename = filename+'.tmp'
    with open(tmpfilename, 'xt', encoding='utf-8') as f:
        f.write(wtext)
    copymode(filename, tmpfilename)
    os.rename(tmpfilename, filename)
    return True
        
variants = (
    {'id': 'qtw',  'name': 'QtWidgets (default)',  'group': 'qtvariant',  'type': 'enum'},
    {'id': 'qtq',  'name': 'QtQuick/QML',          'group': 'qtvariant',  'type': 'enum'},
    {'id': 'ogl',  'name': 'OpenGL (default)',     'group': 'glvariant',  'type': 'enum'},
    {'id': 'es2',  'name': 'OpenGLES2',            'group': 'glvariant',  'type': 'enum'},
    {'id': 'd',    'name': 'debug OpenGL/ES',      'group': 'gldebug',    'type': 'bool'},
    {'id': 'os',   'name': 'offscreen',            'group': 'offscreen',  'type': 'bool'},
    {'id': 'all',  'name': 'all variants',                                'type': None},
)

def accepted_variants():
    return {v['id']: v for v in variants}
    
def variant_string():
    return ', '.join('%s (%s)'%(v['id'],v['name']) for v in variants)
    
def group_variants(variant_ids):
    grouped_variants = {}
    for v in variants:
        if v['type'] == None:
            continue
        values, all_values = grouped_variants.setdefault(v['group'], [[], []])
        if v['type'] == 'bool':
            values.append(None)
            all_values.append(None)
        if v['id'] in variant_ids or 'all' in variant_ids:
            values.append(v['id'])
        all_values.append(v['id'])
    # set default enum variants
    for v in variants:
        if v['type'] != 'enum':
            continue
        values, all_values = grouped_variants[v['group']]
        if not values:
            values.append(v['id'])
    for v in variants:
        with suppress(KeyError):
            g = v['group']
            values, all_values = grouped_variants.pop(g)
            yield g, values, all_values
    




#-*- coding:utf-8 -*-
# cython: profile=False

#  Copyright © 2009-2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=W0614,C0326
# although this file is compiled with Python3 syntax, Cython needs at least division from __future__
from __future__ import print_function, division

# This line makes cython happy
global __name__, __package__    # pylint: disable=W0604
#px/__compiled = True
__compiled = False

#px/from libc.math cimport M_PI, cos, sin
from math import radians, cos, sin

#pxm>IF '[[GLDEBUG]]' == 'gldebug'
#pxd+from [[_debug_VARIANT]] cimport *
#pxm>IF '[[GLDEBUG]]' != 'gldebug'
#pxd+from gl_[[GLVARIANT]] cimport *
#pxm>IF_END

#px:
from pybiklib.debug import DEBUG_LOGGL
try:
    if DEBUG_LOGGL:
        import OpenGL
        OpenGL.FULL_LOGGING = True
    from OpenGL.GLU import * # XXX: without the next line fails with python 3.5
    from OpenGL.GL import *     # pylint: disable=W0614,W0401
except ImportError as e:
    print('The pure Python mode needs PyOpenGL (for Python 3):', e)
    raise SystemExit(1)
import ctypes
from ctypes import sizeof
from OpenGL.raw.GL.VERSION.GL_2_0 import glVertexAttribPointer
#px.


#px/cdef enum: #
if True:
    DEBUG_MSGEXT = 8
#px+cdef long debug
debug = 0
    
def set_debug_flags(module):
    global debug
    if module.DEBUG_MSGEXT: debug |= DEBUG_MSGEXT
    
#px-
NULL = None

#pxd>cdef enum: #
#px-
if True:
    #pxd:
    MAX_TRANSFORMATIONS = 60
    MAX_BLOCKS = 1312
    MAX_FACES = 20
    ATTRIB_LOCATION = 0
    PICKATTRIB_LOCATION = 5
    ATTRIB_CNT = 7
    #px.

#pxd>ctypedef float vec4[4]
#pxd>ctypedef vec4 mat4[4]
#px:
vec4 = lambda: [0.]*4
mat4 = lambda: [[0.]*4 for _i in range(4)]
#px.

#pxm-FUNC PD nogil
def matrix_set_identity(matrix:'mat4 &'):
    matrix[0][0] = 1.; matrix[0][1] = 0.; matrix[0][2] = 0.; matrix[0][3] = 0.
    matrix[1][0] = 0.; matrix[1][1] = 1.; matrix[1][2] = 0.; matrix[1][3] = 0.
    matrix[2][0] = 0.; matrix[2][1] = 0.; matrix[2][2] = 1.; matrix[2][3] = 0.
    matrix[3][0] = 0.; matrix[3][1] = 0.; matrix[3][2] = 0.; matrix[3][3] = 1.
    
#px/cdef struct Block:
class Block:     # pylint: disable=R0903
#px-
    def __init__(self):
        #px/vec4 *transformation
        self.transformation = None
        
        #px/bint in_motion
        self.in_motion = None
        
        #px/int idx_triangles
        self.idx_triangles = None
        #px/int cnt_triangles
        self.cnt_triangles = None
        
#px/cdef struct Cube:
class cube:     # pylint: disable=W0232, R0903
    #px/mat4 transformations[MAX_TRANSFORMATIONS]
    transformations = [[[None]*4, [None]*4, [None]*4, [None]*4] for __t in range(MAX_TRANSFORMATIONS)]
        
    #px+unsigned int number_blocks
    #px/Block blocks[MAX_BLOCKS]
    blocks = [Block() for __block in range(MAX_BLOCKS)]
    
    #px+int cnt_pick
    #px+int idx_debug
    #px+int cnt_debug
    
    #px+GLuint object_location
    #px+GLuint glbuffer
    
    #px+int vertexdatalen
    #px+char *vertexdata
    #px/long attrib_pointers[ATTRIB_CNT]
    attrib_pointers = [None] * ATTRIB_CNT
    
    # animated slice
    #px/mat4 rotation_matrix
    rotation_matrix = mat4()
    
#px+cdef Cube cube


#pxm-FUNC PD
def init_gldraw():
    if debug & DEBUG_MSGEXT:
        print('init module:', __name__)
        print('  from package:', __package__)
        print('  compiled:', __compiled)
        #px+print('  GL-type: [[GLVARIANT]]')
    cube.number_blocks = 0
    cube.cnt_pick = 0
    cube.idx_debug = 0
    cube.cnt_debug = 0
    cube.vertexdatalen = 0
    cube.vertexdata = NULL
    
#pxm-FUNC PD nogil
def sync_block_transformations(blocks:'short *'):
    #px+cdef unsigned int i
    for i in range(cube.number_blocks):
        cube.blocks[i].transformation = cube.transformations[blocks[i]]
        cube.blocks[i].in_motion = False
    
#pxm-FUNC PD nogil
def sync_animation_start(blocks_count:int, blocks:'short *'):
    matrix_set_identity(cube.rotation_matrix)
    #px+cdef int i
    for i in range(blocks_count):
        cube.blocks[blocks[i]].in_motion = True
    
#pxm-FUNC PD nogil
def sync_animation_next(angle:float, rotation_x:float, rotation_y:float, rotation_z:float):
    #px/angle = angle / 180. * M_PI
    angle = radians(angle)
    #px+cdef float x, y, z
    x = rotation_x
    y = rotation_y
    z = rotation_z
    #px+cdef float c, s
    c = cos(angle)
    s = sin(angle)
    
    cube.rotation_matrix[0][0] = x*x*(1-c) + c
    cube.rotation_matrix[0][1] = x*y*(1-c) + z*s
    cube.rotation_matrix[0][2] = x*z*(1-c) - y*s
    cube.rotation_matrix[1][0] = y*x*(1-c) - z*s
    cube.rotation_matrix[1][1] = y*y*(1-c) + c
    cube.rotation_matrix[1][2] = y*z*(1-c) + x*s
    cube.rotation_matrix[2][0] = x*z*(1-c) + y*s
    cube.rotation_matrix[2][1] = y*z*(1-c) - x*s
    cube.rotation_matrix[2][2] = z*z*(1-c) + c
    
#pxm-FUNC P nogil
def matrix_mult(dest:'mat4 &', src1:'mat4 &', src2:'mat4 &'):
    #px+cdef int i, j, k
    #px+cdef float sum_
    for j in range(4):
        for i in range(4):
            sum_ = 0.
            for k in range(4):
                sum_ += src1[k][i] * src2[j][k]
            dest[j][i] = sum_
            
#pxm-FUNC PD nogil
def gl_draw_cube():
    #px+cdef unsigned int i
    #px/cdef mat4 object_matrix
    object_matrix = mat4()
    for i in range(cube.number_blocks):
        if cube.blocks[i].in_motion:
            matrix_mult(object_matrix, cube.rotation_matrix, cube.blocks[i].transformation)
            #px/glUniformMatrix4fv(cube.object_location, 1, GL_FALSE, &object_matrix[0][0])
            glUniformMatrix4fv(cube.object_location, 1, GL_FALSE, object_matrix)
        else:
            #px/glUniformMatrix4fv(cube.object_location, 1, GL_FALSE, &cube.blocks[i].transformation[0][0])
            glUniformMatrix4fv(cube.object_location, 1, GL_FALSE, cube.blocks[i].transformation)
        glDrawArrays(GL_TRIANGLES, cube.blocks[i].idx_triangles, cube.blocks[i].cnt_triangles)
        
#pxm-FUNC PD nogil
def gl_pick_cube():
    glDrawArrays(GL_TRIANGLES, 0, cube.cnt_pick)
    
#pxm-FUNC PD nogil
def gl_init_buffers():
    #px/glGenBuffers(1, &cube.glbuffer)
    cube.glbuffer = glGenBuffers(1)
    
#pxm-FUNC PD nogil
def gl_delete_buffers():
    glBindBuffer(GL_ARRAY_BUFFER, 0)
    #px/glDeleteBuffers(1, &cube.glbuffer)
    glDeleteBuffers(1, [cube.glbuffer])
    cube.glbuffer = 0
    
#pxm-FUNC P nogil
def _gl_enable_pointer(index:'GLuint', size:'GLint', type:'GLenum', normalized:'GLboolean', pointer:'long'):
    #px/glVertexAttribPointer(index, size, type, normalized, 0, <void*>pointer)
    glVertexAttribPointer(index, size, type, normalized, 0, ctypes.cast(pointer, ctypes.c_void_p))
    glEnableVertexAttribArray(index)
    
#pxm-FUNC P nogil
def _gl_disable_pointer(index:'GLuint'):
    #px/glVertexAttribPointer(index, 4, GL_FLOAT, GL_FALSE, 0, NULL)
    glVertexAttribPointer(index, 4, GL_FLOAT, GL_FALSE, 0, ctypes.cast(0, ctypes.c_void_p))
    glDisableVertexAttribArray(index)
    
#pxm-FUNC PD nogil
def sync_blocks(nblocks:int, cnts_block:'short *', idx_debug:int, cnt_debug:int, cnt_pick:int):
    #px+cdef unsigned int idx_block, idx
    cube.number_blocks = nblocks
    
    idx_block = 0
    for idx in range(cube.number_blocks):
        cube.blocks[idx].idx_triangles = idx_block
        cube.blocks[idx].cnt_triangles = cnts_block[idx]
        idx_block += cnts_block[idx]
        
    cube.cnt_pick = cnt_pick
    cube.idx_debug = idx_debug
    cube.cnt_debug = cnt_debug
    
#pxm-FUNC PD nogil
def sync_vertexdata(vertexdatalen:int, vertexdata:'char *', vertexpointers:'long *'):
    cube.vertexdatalen = vertexdatalen
    cube.vertexdata = vertexdata
    cube.attrib_pointers[0] = 0
    cube.attrib_pointers[1] = vertexpointers[0]
    cube.attrib_pointers[2] = vertexpointers[1]
    cube.attrib_pointers[3] = vertexpointers[2]
    cube.attrib_pointers[4] = vertexpointers[3]
    cube.attrib_pointers[5] = vertexpointers[4]
    cube.attrib_pointers[6] = vertexpointers[5]
    
#pxm-FUNC PD nogil
def sync_transformations(transformations_count:int, transformations:'mat4 *'):
    #px+cdef int t
    #px+cdef unsigned int i, j
    for t in range(transformations_count):
        for i in range(4):
            for j in range(4):
                cube.transformations[t][i][j] = float(transformations[t][i][j])
                
#pxm-FUNC PD nogil
def gl_enable_data():
    glBindBuffer(GL_ARRAY_BUFFER, cube.glbuffer)
    glBufferData(GL_ARRAY_BUFFER, cube.vertexdatalen, cube.vertexdata, GL_STATIC_DRAW)
        
    _gl_enable_pointer(ATTRIB_LOCATION,   3, GL_FLOAT, GL_FALSE, cube.attrib_pointers[0])
    _gl_enable_pointer(ATTRIB_LOCATION+1, 3, GL_FLOAT, GL_FALSE, cube.attrib_pointers[1])
    _gl_enable_pointer(ATTRIB_LOCATION+2, 3, GL_UNSIGNED_BYTE, GL_TRUE, cube.attrib_pointers[2])
    _gl_enable_pointer(ATTRIB_LOCATION+3, 2, GL_FLOAT, GL_FALSE, cube.attrib_pointers[3])
    _gl_enable_pointer(ATTRIB_LOCATION+4, 3, GL_FLOAT, GL_FALSE, cube.attrib_pointers[4])
    _gl_enable_pointer(PICKATTRIB_LOCATION, 3, GL_FLOAT, GL_FALSE, cube.attrib_pointers[5])
    _gl_enable_pointer(PICKATTRIB_LOCATION+1, 3, GL_UNSIGNED_BYTE, GL_TRUE, cube.attrib_pointers[6])
    
#pxm-FUNC PD nogil
def gl_disable_data():
    glBindBuffer(GL_ARRAY_BUFFER, 0)
    _gl_disable_pointer(ATTRIB_LOCATION)
    _gl_disable_pointer(ATTRIB_LOCATION+1)
    _gl_disable_pointer(ATTRIB_LOCATION+2)
    _gl_disable_pointer(ATTRIB_LOCATION+3)
    _gl_disable_pointer(ATTRIB_LOCATION+4)
    _gl_disable_pointer(PICKATTRIB_LOCATION)
    _gl_disable_pointer(PICKATTRIB_LOCATION+1)
    
#pxm-FUNC PD nogil
def gl_draw_cube_debug():
    #px/cdef mat4 object_matrix
    object_matrix = mat4()
    matrix_set_identity(object_matrix)
    #px/glUniformMatrix4fv(cube.object_location, 1, GL_FALSE, &object_matrix[0][0])
    glUniformMatrix4fv(cube.object_location, 1, GL_FALSE, object_matrix)
    glDrawArrays(GL_LINES, cube.idx_debug, cube.cnt_debug)
    
#pxm-FUNC PD nogil
def gl_draw_select_debug(selectdata:'GLfloat *', size:'GLsizeiptr', prog_hud:'GLuint'):
    #px+cdef int i, j
    #px+cdef GLintptr offset
    if cube.cnt_debug == 0:
        #XXX: The data used here has fixed size, but may not be initialized yet
        return
    offset = (cube.idx_debug+cube.cnt_debug) * 3 * sizeof(GLfloat)
    #px/glBufferSubData(GL_ARRAY_BUFFER, offset, size, &selectdata[0])
    glBufferSubData(GL_ARRAY_BUFFER, offset, len(selectdata) * sizeof(GLfloat), ArrayDatatype.asArray(selectdata, GL_FLOAT))
    
    glDisable(GL_DEPTH_TEST)
    glDrawArrays(GL_LINES, cube.idx_debug+cube.cnt_debug, 2)
    glUseProgram(prog_hud)
    glDrawArrays(GL_POINTS, cube.idx_debug+cube.cnt_debug+2, 2)
    
#pxm-FUNC PD nogil
def gl_init_object_location(location:'GLuint'):
    cube.object_location = location
    

